import re

import numpy
import pytest

from grunnur import dtypes
from grunnur.dtypes import FieldInfo, _WrappedType
from grunnur.modules import render_with_modules


def test_ctype_builtin():
    assert dtypes.ctype(numpy.int32) == "int"

    with pytest.raises(ValueError, match="object is not a built-in data type"):
        dtypes.ctype(numpy.object_)


def test_is_complex():
    assert dtypes.is_complex(numpy.complex64)
    assert dtypes.is_complex(numpy.complex128)
    assert not dtypes.is_complex(numpy.float64)


def test_is_double():
    assert dtypes.is_double(numpy.float64)
    assert dtypes.is_double(numpy.complex128)
    assert not dtypes.is_double(numpy.complex64)


def test_is_integer():
    assert dtypes.is_integer(numpy.int32)
    assert not dtypes.is_integer(numpy.float32)


def test_is_real():
    assert dtypes.is_real(numpy.float32)
    assert not dtypes.is_real(numpy.complex64)
    assert not dtypes.is_real(numpy.int32)


def test_promote_type():
    assert dtypes._promote_type(numpy.dtype("int8")) == numpy.int32
    assert dtypes._promote_type(numpy.dtype("uint8")) == numpy.uint32
    assert dtypes._promote_type(numpy.dtype("float16")) == numpy.float32
    assert dtypes._promote_type(numpy.dtype("csingle")) == numpy.complex64
    assert dtypes._promote_type(numpy.dtype("int32")) == numpy.int32
    assert dtypes._promote_type(numpy.dtype("int64")) == numpy.int64


def test_result_type():
    assert dtypes.result_type(numpy.int32, numpy.float32) == numpy.float64


def test_min_scalar_type():
    assert dtypes.min_scalar_type(1) == numpy.uint32
    assert dtypes.min_scalar_type(-1) == numpy.int32
    assert dtypes.min_scalar_type(1.0) == numpy.float32
    assert dtypes.min_scalar_type(1 + 2j) == numpy.complex64


def test_complex_for():
    assert dtypes.complex_for(numpy.float32) == numpy.complex64
    assert dtypes.complex_for(numpy.float64) == numpy.complex128
    with pytest.raises(ValueError, match="complex64 does not have a corresponding complex type"):
        assert dtypes.complex_for(numpy.complex64)
    with pytest.raises(ValueError, match="int32 does not have a corresponding complex type"):
        assert dtypes.complex_for(numpy.int32)


def test_real_for():
    assert dtypes.real_for(numpy.complex64) == numpy.float32
    assert dtypes.real_for(numpy.complex128) == numpy.float64
    with pytest.raises(ValueError, match="float32 does not have a corresponding real type"):
        assert dtypes.real_for(numpy.float32)
    with pytest.raises(ValueError, match="int32 does not have a corresponding real type"):
        assert dtypes.real_for(numpy.int32)


def test_complex_ctr():
    assert dtypes.complex_ctr(numpy.complex64) == "COMPLEX_CTR(float2)"


def test_c_constant():
    # scalar values
    assert dtypes.c_constant(1) == "1"
    assert dtypes.c_constant(numpy.uint64(1)) == "1UL"
    assert dtypes.c_constant(numpy.int64(-1)) == "-1L"
    assert dtypes.c_constant(numpy.float64(1.0)) == "1.0"
    assert dtypes.c_constant(numpy.float32(1.0)) == "1.0f"
    assert dtypes.c_constant(numpy.complex64(1 + 2j)) == "COMPLEX_CTR(float2)(1.0f, 2.0f)"
    assert dtypes.c_constant(numpy.complex128(1 + 2j)) == "COMPLEX_CTR(double2)(1.0, 2.0)"

    # array
    assert dtypes.c_constant(numpy.array([1, 2, 3], numpy.float32)) == "{1.0f, 2.0f, 3.0f}"

    # struct type
    dtype = numpy.dtype([("val1", numpy.int32), ("val2", numpy.float32)])
    val = numpy.empty((), dtype)
    val["val1"] = 1
    val["val2"] = 2
    assert dtypes.c_constant(val) == "{1, 2.0f}"

    # custom dtype
    assert dtypes.c_constant(1, numpy.float32) == "1.0f"

    message = r"Cannot render a value of type <class 'numpy.str_'> as a C constant"
    with pytest.raises(TypeError, match=message):
        dtypes.c_constant(numpy.array(["a", "b"]))


def test_align_builtin():
    dtype = numpy.dtype("int32")
    assert dtypes.align(dtype) == dtype


def test_wrap_builtin():
    dtype = numpy.dtype("int32")
    ref = _WrappedType(
        dtype=dtype,
        alignment=4,
        wrapped_fields={},
        explicit_alignment=None,
        explicit_field_alignments={},
    )
    assert _WrappedType.wrap(dtype) == ref


def test_align_array():
    dtype = numpy.dtype("int32")
    dtype_arr = numpy.dtype((dtype, 3))
    assert dtypes.align(dtype_arr) == dtype_arr


def test_wrap_array():
    dtype = numpy.dtype("int32")
    dtype_arr = numpy.dtype((dtype, 3))
    ref = _WrappedType(
        dtype=dtype_arr,
        alignment=4,
        wrapped_fields={},
        explicit_alignment=None,
        explicit_field_alignments={},
    )
    assert _WrappedType.wrap(dtype_arr) == ref


def test_align_sets_aligned_attribute():
    dtype = numpy.dtype(dict(names=["x", "y", "z"], formats=[numpy.int8, numpy.int16, numpy.int32]))
    res = dtypes.align(dtype)

    assert res.isalignedstruct
    assert dict(res.fields) == dict(x=(numpy.int8, 0), y=(numpy.int16, 2), z=(numpy.int32, 4))
    assert res.itemsize == 8
    assert res.alignment == 4


def test_wrap_aligned_struct():
    dtype = numpy.dtype(
        dict(
            names=["x", "y", "z"],
            formats=[numpy.int8, numpy.int16, numpy.int32],
            offsets=[0, 2, 4],
            itemsize=8,
            aligned=True,
        )
    )

    wt_x = _WrappedType.wrap(numpy.dtype("int8"))
    wt_y = _WrappedType.wrap(numpy.dtype("int16"))
    wt_z = _WrappedType.wrap(numpy.dtype("int32"))
    ref = _WrappedType(
        dtype=dtype,
        alignment=4,
        explicit_alignment=None,
        wrapped_fields=dict(x=wt_x, y=wt_y, z=wt_z),
        explicit_field_alignments=dict(x=None, y=None, z=None),
    )
    assert _WrappedType.wrap(dtype) == ref


def test_align_ignores_offsets_and_itemsize():
    dtype = numpy.dtype(
        dict(
            names=["x", "y", "z"],
            formats=[numpy.int8, numpy.int16, numpy.int32],
            offsets=[0, 4, 16],
            itemsize=32,
            aligned=True,
        )
    )

    # Aligning will ignore all existing offets and itemsizes,
    # so the itemsize will be reset to 8 and alignment to 4.
    res = dtypes.align(dtype)

    assert res.isalignedstruct
    assert dict(res.fields) == dict(x=(numpy.int8, 0), y=(numpy.int16, 2), z=(numpy.int32, 4))
    assert res.itemsize == 8
    assert res.alignment == 4


def test_wrap_custom_offsets_and_itemsize():
    dtype = numpy.dtype(
        dict(
            names=["x", "y", "z"],
            formats=[numpy.int8, numpy.int16, numpy.int32],
            offsets=[0, 4, 16],
            itemsize=32,
            aligned=True,
        )
    )

    wt_x = _WrappedType.wrap(numpy.dtype("int8"))
    wt_y = _WrappedType.wrap(numpy.dtype("int16"))
    wt_z = _WrappedType.wrap(numpy.dtype("int32"))

    # The offsets & itemsize are possible to achieve with alignments,
    # so this won't fail.
    res = _WrappedType.wrap(dtype)
    ref = _WrappedType(
        dtype=dtype,
        alignment=16,
        explicit_alignment=None,
        wrapped_fields=dict(x=wt_x, y=wt_y, z=wt_z),
        explicit_field_alignments=dict(x=None, y=4, z=16),
    )
    assert res == ref


def test_wrap_invalid_itemsize():
    dtype = numpy.dtype(
        dict(
            names=["x", "y", "z"],
            formats=[numpy.int8, numpy.int16, numpy.int32],
            offsets=[0, 2, 4],
            # This itemsize is not the implicit one based on the field sizes and alignments,
            # and is not a power of 2, so it cannot be caused by an explicit alignment.
            # An error should be raised.
            itemsize=20,
            aligned=True,
        )
    )

    with pytest.raises(
        ValueError, match="An itemsize that requires an explicit alignment must be a power of 2"
    ):
        _WrappedType.wrap(dtype)


def test_align_nested():
    dtype_nested = numpy.dtype(dict(names=["val1", "pad"], formats=[numpy.int8, numpy.int8]))

    dtype = numpy.dtype(
        dict(
            names=["pad", "struct_arr", "regular_arr"],
            formats=[numpy.int32, numpy.dtype((dtype_nested, 2)), numpy.dtype((numpy.int16, 3))],
        )
    )

    dtype_ref = numpy.dtype(
        dict(
            names=["pad", "struct_arr", "regular_arr"],
            formats=[numpy.int32, (dtype_nested, (2,)), (numpy.int16, (3,))],
            offsets=[0, 4, 8],
            itemsize=16,
            aligned=True,
        )
    )

    dtype_aligned = dtypes.align(dtype)

    assert dtype_aligned == dtype_ref


def test_align_nested_ignores_itemsize():
    dtype_int3 = numpy.dtype(
        dict(names=["x"], formats=[(numpy.int32, 3)], itemsize=16, aligned=True)
    )
    dtype = numpy.dtype(
        dict(names=["x", "y", "z"], formats=[numpy.int32, dtype_int3, numpy.int32], aligned=True)
    )

    # Alignment will ignore the existing itemsize and pack the components tighter.
    dtype_int3_ref = numpy.dtype(
        dict(names=["x"], formats=[(numpy.int32, 3)], itemsize=12, aligned=True)
    )
    dtype_ref = numpy.dtype(
        dict(
            names=["x", "y", "z"],
            formats=[numpy.int32, dtype_int3_ref, numpy.int32],
            offsets=[0, 4, 16],
            itemsize=20,
            aligned=True,
        )
    )

    dtype_aligned = dtypes.align(dtype)

    assert dtype_aligned.isalignedstruct
    assert dtype_aligned == dtype_ref


def test_lcm():
    assert dtypes._lcm(10) == 10
    assert dtypes._lcm(15, 20) == 60
    assert dtypes._lcm(16, 32, 24) == 96


def test_find_minimum_alignment():
    # simple case: base alignment is enough because 12 is the next multiple of 4 after 9
    assert dtypes._find_minimum_alignment(12, 4, 9) == 4
    # the next multiple of 4 is 12, but we want offset 16 - this means we need to set
    # the alignment equal to 8, because 16 is the next multiple of 8 after 9.
    assert dtypes._find_minimum_alignment(16, 4, 9) == 8

    # incorrect offset (not a multiple of the base alignment)
    message = re.escape("Field offset (13) must be a multiple of the base alignment (4)")
    with pytest.raises(ValueError, match=message):
        dtypes._find_minimum_alignment(13, 4, 9)

    # offset too large and not a power of 2 - cannot achieve that with alignment only,
    # will need explicit padding
    message = (
        "Could not find a suitable alignment for the field at offset 24; "
        "consider adding explicit padding"
    )
    with pytest.raises(ValueError, match=message):
        dtypes._find_minimum_alignment(24, 4, 9)


def test_ctype_struct():
    dtype = dtypes.align(numpy.dtype([("val1", numpy.int32), ("val2", numpy.float32)]))
    ctype = dtypes.ctype(dtype)
    src = render_with_modules("${ctype}", render_globals=dict(ctype=ctype)).strip()

    assert src == (
        "typedef struct _mod__module_0__ {\n"
        "    int  val1;\n"
        "    float  val2;\n"
        "}  _mod__module_0_;\n\n\n"
        "_mod__module_0_"
    )


def test_ctype_struct_nested():
    dtype_nested = numpy.dtype(dict(names=["val1", "pad"], formats=[numpy.int8, numpy.int8]))

    dtype = numpy.dtype(
        dict(
            names=["pad", "struct_arr", "regular_arr"],
            formats=[numpy.int32, numpy.dtype((dtype_nested, 2)), numpy.dtype((numpy.int16, 3))],
        )
    )

    dtype = dtypes.align(dtype)
    ctype = dtypes.ctype(dtype)
    src = render_with_modules("${ctype}", render_globals=dict(ctype=ctype)).strip()

    assert src == (
        "typedef struct _mod__module_1__ {\n"
        "    char  val1;\n"
        "    char  pad;\n"
        "}  _mod__module_1_;\n\n\n"
        "typedef struct _mod__module_0__ {\n"
        "    int  pad;\n"
        "    _mod__module_1_  struct_arr[2];\n"
        "    short  regular_arr[3];\n"
        "}  _mod__module_0_;\n\n\n"
        "_mod__module_0_"
    )


def test_ctype_struct_aligned():
    dtype = numpy.dtype(
        dict(
            names=["x", "y", "z"],
            formats=[numpy.int8, numpy.int16, numpy.int32],
            offsets=[0, 4, 16],
            itemsize=64,
            aligned=True,
        )
    )
    ctype = dtypes.ctype(dtype)
    src = render_with_modules("${ctype}", render_globals=dict(ctype=ctype)).strip()
    assert src == (
        "typedef struct _mod__module_0__ {\n"
        "    char  x;\n"
        "    short ALIGN(4) y;\n"
        "    int ALIGN(16) z;\n"
        "} ALIGN(64) _mod__module_0_;\n\n\n"
        "_mod__module_0_"
    )


def test_ctype_checks_alignment():
    dtype = numpy.dtype(dict(names=["x", "y", "z"], formats=[numpy.int8, numpy.int16, numpy.int32]))
    message = re.escape(
        "Failed to find alignment for field `y`: "
        "Field offset (1) must be a multiple of the base alignment (2)."
    )
    with pytest.raises(ValueError, match=message):
        dtypes.ctype(dtype)


def test_ctype_for_array():
    dtype = numpy.dtype((numpy.int32, 3))
    with pytest.raises(ValueError, match="The data type cannot be an array"):
        dtypes.ctype(dtype)


def test_flatten_dtype():
    dtype_nested = numpy.dtype(dict(names=["val1", "pad"], formats=[numpy.int8, numpy.int8]))

    dtype = numpy.dtype(
        dict(
            names=["pad", "struct_arr", "regular_arr"],
            formats=[numpy.int32, numpy.dtype((dtype_nested, 2)), numpy.dtype((numpy.int16, 3))],
        )
    )

    res = dtypes.flatten_dtype(dtype)
    ref = [
        FieldInfo(path=["pad"], dtype=numpy.dtype("int32"), offset=0),
        FieldInfo(path=["struct_arr", 0, "val1"], dtype=numpy.dtype("int8"), offset=4),
        FieldInfo(path=["struct_arr", 0, "pad"], dtype=numpy.dtype("int8"), offset=5),
        FieldInfo(path=["struct_arr", 1, "val1"], dtype=numpy.dtype("int8"), offset=6),
        FieldInfo(path=["struct_arr", 1, "pad"], dtype=numpy.dtype("int8"), offset=7),
        FieldInfo(path=["regular_arr", 0], dtype=numpy.dtype("int16"), offset=8),
        FieldInfo(path=["regular_arr", 1], dtype=numpy.dtype("int16"), offset=10),
        FieldInfo(path=["regular_arr", 2], dtype=numpy.dtype("int16"), offset=12),
    ]

    assert res == ref


def test_c_path():
    field_info = FieldInfo(path=["struct_arr", 0, "val1"], dtype=numpy.int8, offset=0)
    assert field_info.c_path == ".struct_arr[0].val1"


def test_c_path_primitive_type():
    flat_dtype = dtypes.flatten_dtype(numpy.int32)
    assert flat_dtype[0].c_path == ""


def test_extract_field():
    dtype_nested = numpy.dtype(dict(names=["val1", "pad"], formats=[numpy.int8, numpy.int8]))

    dtype = numpy.dtype(
        dict(
            names=["pad", "struct_arr", "regular_arr"],
            formats=[numpy.int32, numpy.dtype((dtype_nested, 2)), numpy.dtype((numpy.int16, 3))],
        )
    )

    a = numpy.empty(16, dtype)
    a["struct_arr"]["val1"][:, 1] = numpy.arange(16)
    assert (dtypes.extract_field(a, ["struct_arr", 1, "val1"]) == numpy.arange(16)).all()

    b = numpy.empty(16, dtype_nested)
    b["val1"] = numpy.arange(16)
    assert (dtypes.extract_field(b, ["val1"]) == numpy.arange(16)).all()



import io

import pytest
import pytest_grunnur.plugin
from pytest_grunnur import get_devices, get_multi_device_sets

# Cannot just use the plugin directly since it is loaded before the coverage plugin,
# and all the things the plugin uses from `grunnur` get marked as not covered.
from pytest_grunnur.plugin import context, device

from grunnur import API, Context, Device, cuda_api_id, opencl_api_id
from grunnur.adapter_base import APIID
from grunnur.api import all_api_ids
from grunnur.testing import MockBackendFactory, MockPyCUDA, MockPyOpenCL
from grunnur.virtual_alloc import TrivialManager, ZeroOffsetManager


def pytest_addoption(parser):
    # Manually adding options from `pytest_grunnur` (see the note in the imports)
    pytest_grunnur.plugin.pytest_addoption(parser)


def pytest_report_header(config):
    # Manually adding the header from `pytest_grunnur` (see the note in the imports)
    pytest_grunnur.plugin.pytest_report_header(config)


@pytest.fixture
def mock_backend_factory(monkeypatch):
    return MockBackendFactory(monkeypatch)


@pytest.fixture(params=all_api_ids(), ids=lambda api_id: f"mock_backend_{api_id.shortcut}")
def mock_backend(request, mock_backend_factory):
    return mock_backend_factory.mock(request.param)


@pytest.fixture
def mock_backend_pyopencl(mock_backend_factory):
    return mock_backend_factory.mock_pyopencl()


@pytest.fixture
def mock_backend_pycuda(mock_backend_factory):
    return mock_backend_factory.mock_pycuda()


def make_context(backend, devices_num):
    api_id = backend.api_id
    backend.add_devices(["Device" + str(i) for i in range(devices_num)])
    api = API.from_api_id(api_id)
    return Context.from_criteria(api, devices_num=devices_num)


@pytest.fixture(params=all_api_ids(), ids=lambda api_id: f"mock_{api_id.shortcut}")
def mock_context(request, mock_backend_factory):
    backend = mock_backend_factory.mock(request.param)
    return make_context(backend, 1)


@pytest.fixture
def mock_context_pycuda(mock_backend_pycuda):
    return make_context(mock_backend_pycuda, 1)


@pytest.fixture(params=all_api_ids(), ids=lambda api_id: f"mock_{api_id.shortcut}_4dev")
def mock_4_device_context(request, mock_backend_factory):
    backend = mock_backend_factory.mock(request.param)
    return make_context(backend, 4)


@pytest.fixture
def mock_4_device_context_pyopencl(mock_backend_pyopencl):
    return make_context(mock_backend_pyopencl, 4)


@pytest.fixture
def mock_or_real_context(request, monkeypatch):
    # Since `py.test` does not support concatenating fixtures,
    # we concatenate real contexts and mocked contexts manually.
    # If there are no devices available, we need it to be noticeable that some tests are skipped,
    # hence passing `None` and calling `skip()` explicitly in this case.
    value = request.param
    if value is None:
        pytest.skip("No GPGPU devices available")
    elif isinstance(value, Device):
        yield (Context.from_devices([value]), False)
    elif isinstance(value, APIID):
        factory = MockBackendFactory(monkeypatch)
        backend = factory.mock(value)
        yield (make_context(backend, 1), True)
    else:
        raise TypeError


@pytest.fixture
def mock_or_real_multi_device_context(request, monkeypatch):
    # Same as `mock_or_real_context` above, but for 2-device contexts.
    value = request.param
    if value is None:
        pytest.skip("Could not find 2 suitable GPGPU devices on the same platform")
    elif isinstance(value, list):
        yield (Context.from_devices(value), False)
    elif isinstance(value, APIID):
        factory = MockBackendFactory(monkeypatch)
        backend = factory.mock(value)
        yield (make_context(backend, 2), True)
    else:
        raise TypeError


class MockStdin:
    def __init__(self):
        self.stream = io.StringIO()
        self.lines = 0

    def line(self, s):
        pos = self.stream.tell()  # preserve the current read pointer
        self.stream.seek(0, io.SEEK_END)
        self.stream.write(s + "\n")
        self.stream.seek(pos)
        self.lines += 1

    def readline(self):
        assert self.lines > 0
        self.lines -= 1
        return self.stream.readline()

    def empty(self):
        return self.lines == 0


@pytest.fixture
def mock_stdin(monkeypatch):
    mock = MockStdin()
    monkeypatch.setattr("sys.stdin", mock)
    yield mock
    assert mock.empty()


@pytest.fixture(params=[TrivialManager, ZeroOffsetManager], ids=["trivial", "zero_offset"])
def valloc_cls(request):
    return request.param


def _mock_or_real_context_make_id(value):
    if value is None:
        return "no_real_devices"
    if isinstance(value, Device):
        return value.shortcut
    if isinstance(value, APIID):
        return f"mock_{value.shortcut}"
    raise TypeError


def _mock_or_real_multi_device_context_make_id(value):
    if value is None:
        return "no_2_real_devices"
    if isinstance(value, list):
        return "+".join(device.shortcut for device in value)
    if isinstance(value, APIID):
        return f"mock_{value.shortcut}_2dev"
    raise TypeError


def pytest_generate_tests(metafunc):
    # Manually adding the fixtures from `pytest_grunnur` (see the note in the imports)
    pytest_grunnur.plugin.pytest_generate_tests(metafunc)

    api_ids = all_api_ids()
    devices = get_devices(metafunc.config)
    device_sets = get_multi_device_sets(metafunc.config)

    if "mock_or_real_context" in metafunc.fixturenames:
        values = (devices if len(devices) > 0 else [None]) + api_ids
        ids = [_mock_or_real_context_make_id(value) for value in values]
        metafunc.parametrize("mock_or_real_context", values, ids=ids, indirect=True)

    if "mock_or_real_multi_device_context" in metafunc.fixturenames:
        values = (device_sets if len(device_sets) > 0 else [None]) + api_ids
        ids = [_mock_or_real_multi_device_context_make_id(value) for value in values]
        metafunc.parametrize("mock_or_real_multi_device_context", values, ids=ids, indirect=True)



# This is a testing module, so it has a lot of assertions and private attribute access.
# ruff: noqa: S101, SLF001

from __future__ import annotations

import os.path
import weakref
from abc import ABC, abstractmethod
from enum import Enum
from functools import lru_cache
from pathlib import Path
from tempfile import mkdtemp
from typing import TYPE_CHECKING, Any, Protocol, cast

import numpy

from .. import cuda_api_id
from ..utils import prod
from .mock_base import MockKernel, MockSource

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Sequence


class ConvertibleToInt(Protocol):
    def __int__(self) -> int:  # pragma: no cover
        ...


class MockPyCUDA:
    def __init__(self, cuda_version: tuple[int, int, int] = (10, 0, 0)):
        self.pycuda_driver = MockPyCUDADriverModule(self, cuda_version)
        self.pycuda_compiler = MockPyCUDACompilerModule(self)
        self.pycuda_tools = MockPyCUDATools()

        self.device_infos: list[PyCUDADeviceInfo] = []
        self._context_stack: list[weakref.ReferenceType[BaseContext]] = []

        # Since we need to cast DeviceAllocation objects to integers (to add offsets),
        # there is no way to use a mock allocation object to track that.
        # Instead, we have to use recognizable integers as "addresses" and check the validity of
        # allocations using a kind of a fuzzy match database.
        # Should work for testing purposes as long as we use small offsets,
        # and other integer parameters don't fall in the "address" range.
        self._allocation_start = 2**30
        self._allocation_step = 2**16
        self._allocation_idx = 0
        self._allocations: dict[int, tuple[int, weakref.ReferenceType[BaseContext], bytes]] = {}

        self.api_id = cuda_api_id()

    def add_devices(self, device_infos: Sequence[str | PyCUDADeviceInfo]) -> None:
        assert len(self.device_infos) == 0
        for device_info in device_infos:
            if isinstance(device_info, str):
                typed_device_info = PyCUDADeviceInfo(name=device_info)
            else:
                typed_device_info = device_info
            self.device_infos.append(typed_device_info)

    def push_context(self, context: BaseContext) -> None:
        if self.is_stacked(context):
            raise ValueError("The given context is already in the context stack")
        self._context_stack.append(weakref.ref(context))

    def pop_context(self) -> None:
        self._context_stack.pop()

    def current_context(self) -> BaseContext:
        context = self._context_stack[-1]()
        assert context is not None
        return context

    def is_stacked(self, context: BaseContext) -> bool:
        return any(stacked_context_ref() == context for stacked_context_ref in self._context_stack)

    def allocate(self, size: int) -> tuple[int, int]:
        assert size <= self._allocation_step
        idx = self._allocation_idx
        self._allocation_idx += 1
        address = self._allocation_start + self._allocation_step * idx
        self._allocations[idx] = (size, self._context_stack[-1], b"\xef" * size)
        return idx, address

    def get_allocation_buffer(self, idx: int, offset: int, region_size: int) -> bytes:
        size, context, buf = self._allocations[idx]
        return buf[offset : offset + region_size]

    def set_allocation_buffer(self, idx: int, offset: int, data: bytes) -> None:
        size, context, buf = self._allocations[idx]
        self._allocations[idx] = size, context, buf[:offset] + data + buf[offset + len(data) :]

    def allocation_count(self) -> int:
        return len(self._allocations)

    def free_allocation(self, idx: int) -> None:
        del self._allocations[idx]

    def check_allocation(self, address: ConvertibleToInt) -> tuple[int, int, int]:
        # will work as long as we don't have offsets larger than `_allocation_step`
        idx = (int(address) - self._allocation_start) // self._allocation_step
        offset = (int(address) - self._allocation_start) % self._allocation_step

        if idx not in self._allocations:
            raise RuntimeError("A previously freed allocation or an incorrect address is used")

        size, context_ref, _buffer = self._allocations[idx]

        context = context_ref()
        assert context is not None

        if context != self.current_context():
            raise RuntimeError(
                "Trying to access an allocation from a device different to where it was created"
            )

        return idx, offset, size - offset


class PycudaCompileError(Exception):
    pass


class MockPyCUDACompilerModule:
    def __init__(self, backend: MockPyCUDA):
        self.SourceModule = make_source_module_class(backend)


class BaseSourceModule(ABC):
    _context: BaseContext

    @abstractmethod
    def __init__(
        self,
        src: MockSource,
        *,
        no_extern_c: bool = False,
        options: Sequence[str] | None = None,
        keep: bool = False,
        cache_dir: str | None = None,
    ): ...

    @abstractmethod
    def test_get_options(self) -> Sequence[str] | None: ...

    @abstractmethod
    def get_function(self, name: str) -> BaseFunction: ...

    @abstractmethod
    def get_global(self, name: str) -> tuple[BaseDeviceAllocation, int]: ...


@lru_cache
def make_source_module_class(backend: MockPyCUDA) -> type[BaseSourceModule]:
    backend_ref = weakref.ref(backend)

    class SourceModule(BaseSourceModule):
        _backend_ref = backend_ref

        def __init__(
            self,
            src: MockSource,
            *,
            no_extern_c: bool = False,
            options: Sequence[str] | None = None,
            keep: bool = False,
            cache_dir: str | None = None,
        ):
            assert isinstance(src, MockSource)
            assert isinstance(no_extern_c, bool)
            assert options is None or all(isinstance(option, str) for option in options)
            assert isinstance(keep, bool)

            if src.should_fail:
                raise PycudaCompileError

            backend = self._backend_ref()
            assert backend is not None
            self._context = backend.current_context()

            function_cls = backend.pycuda_driver.Function
            self._kernels = {kernel.name: function_cls(self, kernel) for kernel in src.kernels}
            self._constant_mem = src.constant_mem

            self._options = options

            # See the note in compile_single_device(). Apparently that's how PyCUDA operates.
            if keep and cache_dir is not None:
                temp_dir = Path(mkdtemp())
                temp_file = temp_dir / "kernel.cu"
                with temp_file.open("w") as f:
                    f.write(str(src))
                print(f"*** compiler output in {temp_dir}")  # noqa: T201

        def test_get_options(self) -> Sequence[str] | None:
            return self._options

        def get_function(self, name: str) -> BaseFunction:
            return self._kernels[name]

        def get_global(self, name: str) -> tuple[BaseDeviceAllocation, int]:
            size = self._constant_mem[name]
            backend = self._backend_ref()
            assert backend is not None
            alloc = backend.pycuda_driver.DeviceAllocation._allocate(size)
            return alloc, size

    return SourceModule


class MemAttachFlags(Enum):
    GLOBAL = 1


class FunctionAttribute(Enum):
    MAX_THREADS_PER_BLOCK = 0


class MockPyCUDADriverModule:
    Function: type[BaseFunction]

    DeviceAllocation: type[BaseDeviceAllocation]

    def __init__(self, backend: MockPyCUDA, cuda_version: tuple[int, int, int]):
        self._backend_ref = weakref.ref(backend)
        self._version = cuda_version

        self.Device = make_device_class(backend)
        self.Stream = make_stream_class(backend)
        self.Context = make_context_class(backend)
        self.DeviceAllocation = make_device_allocation_class(backend)
        self.Function = make_function_class(backend)

        self.CompileError = PycudaCompileError

        self.mem_attach_flags = MemAttachFlags
        self.function_attribute = FunctionAttribute

    def get_version(self) -> tuple[int, int, int]:
        return self._version

    def mem_alloc(self, size: int) -> BaseDeviceAllocation:
        return self.DeviceAllocation._allocate(size)

    def memcpy_htod(
        self, dest: BaseDeviceAllocation, src: numpy.ndarray[Any, numpy.dtype[Any]]
    ) -> None:
        self.memcpy_htod_async(dest, src)

    def memcpy_htod_async(
        self,
        dest: BaseDeviceAllocation,
        src: numpy.ndarray[Any, numpy.dtype[Any]],
        stream: BaseStream | None = None,
    ) -> None:
        backend = self._backend_ref()
        assert backend is not None
        current_context = backend.current_context()
        assert isinstance(src, numpy.ndarray)
        dest = self.DeviceAllocation._from_memcpy_arg(dest)
        if stream is not None:
            assert stream._context == current_context
        assert dest._size >= src.size * src.dtype.itemsize
        dest._set(src)

    def memcpy_dtoh(
        self, dest: numpy.ndarray[Any, numpy.dtype[Any]], src: BaseDeviceAllocation
    ) -> None:
        self.memcpy_dtoh_async(dest, src)

    def memcpy_dtoh_async(
        self,
        dest: numpy.ndarray[Any, numpy.dtype[Any]],
        src: BaseDeviceAllocation,
        stream: BaseStream | None = None,
    ) -> None:
        backend = self._backend_ref()
        assert backend is not None
        current_context = backend.current_context()
        assert isinstance(dest, numpy.ndarray)
        src = self.DeviceAllocation._from_memcpy_arg(src)
        if stream is not None:
            assert stream._context == current_context
        assert src._size >= dest.size * dest.dtype.itemsize
        src._get(dest)

    def memcpy_dtod(self, dest: BaseDeviceAllocation, src: BaseDeviceAllocation, size: int) -> None:
        self.memcpy_dtod_async(dest, src, size)

    def memcpy_dtod_async(
        self,
        dest: BaseDeviceAllocation,
        src: BaseDeviceAllocation,
        size: int,
        stream: BaseStream | None = None,
    ) -> None:
        backend = self._backend_ref()
        assert backend is not None
        current_context = backend.current_context()
        dest = self.DeviceAllocation._from_memcpy_arg(dest)
        src = self.DeviceAllocation._from_memcpy_arg(src)
        if stream is not None:
            assert stream._context == current_context
        assert dest._size >= size
        assert src._size >= size
        dest._set(src)

    def pagelocked_empty(
        self, shape: Sequence[int], dtype: numpy.dtype[Any]
    ) -> numpy.ndarray[Any, numpy.dtype[Any]]:
        return numpy.empty(shape, dtype)


class PyCUDADeviceInfo:
    def __init__(
        self,
        name: str = "DefaultDeviceName",
        max_threads_per_block: int = 1024,
        max_block_dim_x: int = 1024,
        max_block_dim_y: int = 1024,
        max_block_dim_z: int = 64,
        max_grid_dim_x: int = 2**32 - 1,
        max_grid_dim_y: int = 2**32 - 1,
        max_grid_dim_z: int = 65536,
        warp_size: int = 32,
        max_shared_memory_per_block: int = 64 * 1024,
        multiprocessor_count: int = 8,
        compute_capability: int = 5,
    ):
        self.name = name
        self.max_threads_per_block = max_threads_per_block
        self.max_block_dim_x = max_block_dim_x
        self.max_block_dim_y = max_block_dim_y
        self.max_block_dim_z = max_block_dim_z
        self.max_grid_dim_x = max_grid_dim_x
        self.max_grid_dim_y = max_grid_dim_y
        self.max_grid_dim_z = max_grid_dim_z
        self.warp_size = warp_size
        self.max_shared_memory_per_block = max_shared_memory_per_block
        self.multiprocessor_count = multiprocessor_count
        self.compute_capability = compute_capability


class BaseDevice(ABC):
    max_grid_dim_x: int
    max_grid_dim_y: int
    max_grid_dim_z: int
    max_block_dim_x: int
    max_block_dim_y: int
    max_block_dim_z: int

    @abstractmethod
    def __init__(self, device_idx: int): ...

    @abstractmethod
    def name(self) -> str: ...

    @abstractmethod
    def compute_capability(self) -> tuple[int, int]: ...

    @abstractmethod
    def make_context(self) -> BaseContext: ...

    @staticmethod
    @abstractmethod
    def count() -> int: ...


# We need a device class that is an actual class
# (so that `type()` works on the results of its `__call__()`),
# but at the same time retains a reference to the backend object used to create it
# (so that we can control the properties of the mock).
@lru_cache
def make_device_class(backend: MockPyCUDA) -> type[BaseDevice]:
    backend_ref = weakref.ref(backend)

    class Device(BaseDevice):
        _backend_ref = backend_ref

        def __init__(self, device_idx: int):
            backend = Device._backend_ref()
            assert backend is not None
            device_info = backend.device_infos[device_idx]

            self._device_idx = device_idx
            self._name = device_info.name
            self.max_threads_per_block = device_info.max_threads_per_block

            self.max_block_dim_x = device_info.max_block_dim_x
            self.max_block_dim_y = device_info.max_block_dim_y
            self.max_block_dim_z = device_info.max_block_dim_z

            self.max_grid_dim_x = device_info.max_grid_dim_x
            self.max_grid_dim_y = device_info.max_grid_dim_y
            self.max_grid_dim_z = device_info.max_grid_dim_z

            self.warp_size = device_info.warp_size
            self.max_shared_memory_per_block = device_info.max_shared_memory_per_block
            self.multiprocessor_count = device_info.multiprocessor_count

            self._compute_capability = device_info.compute_capability

        def name(self) -> str:
            return self._name

        def compute_capability(self) -> tuple[int, int]:
            return (self._compute_capability, 0)

        def make_context(self) -> BaseContext:
            backend = self._backend_ref()
            assert backend is not None
            context = backend.pycuda_driver.Context(self._device_idx)
            context.push()
            return context

        def __eq__(self, other: object) -> bool:
            assert isinstance(other, Device)
            return self._device_idx == other._device_idx

        def __hash__(self) -> int:
            return hash((type(self), self._device_idx))

        @staticmethod
        def count() -> int:
            backend = Device._backend_ref()
            assert backend is not None
            return len(backend.device_infos)

    return Device


class BaseContext(ABC):
    _device_idx: int

    @abstractmethod
    def __init__(self, device_idx: int): ...

    @staticmethod
    @abstractmethod
    def pop() -> None: ...

    @staticmethod
    @abstractmethod
    def get_device() -> BaseDevice: ...

    @abstractmethod
    def push(self) -> None: ...

    @abstractmethod
    def is_stacked(self) -> bool: ...


@lru_cache
def make_context_class(backend: MockPyCUDA) -> type[BaseContext]:
    class Context(BaseContext):
        # Since we need the backend in __del__(),
        # we want to make sure that it is alive as long as a this object is alive.
        _backend = backend

        def __init__(self, device_idx: int):
            self._device_idx = device_idx

        @staticmethod
        def pop() -> None:
            Context._backend.pop_context()

        @staticmethod
        def get_device() -> BaseDevice:
            backend = Context._backend
            return backend.pycuda_driver.Device(backend.current_context()._device_idx)

        def push(self) -> None:
            self._backend.push_context(self)

        def is_stacked(self) -> bool:
            return self._backend.is_stacked(self)

        def __del__(self) -> None:
            # Sanity check
            message = "Context was deleted while still being in the context stack"
            assert not self.is_stacked(), message

    return Context


class BaseStream(ABC):
    _context: BaseContext

    @abstractmethod
    def synchronize(self) -> None: ...


@lru_cache
def make_stream_class(backend: MockPyCUDA) -> type[BaseStream]:
    backend_ref = weakref.ref(backend)

    class Stream(BaseStream):
        _backend_ref = backend_ref

        def __init__(self) -> None:
            backend = self._backend_ref()
            assert backend is not None
            self._context = backend.current_context()

        def synchronize(self) -> None:
            backend = self._backend_ref()
            assert backend is not None
            assert self._context == backend.current_context()

    return Stream


class BaseDeviceAllocation(ABC):
    _context: BaseContext
    _idx: int
    _offset: int
    _size: int

    @classmethod
    @abstractmethod
    def _allocate(cls, size: int) -> BaseDeviceAllocation: ...

    @classmethod
    @abstractmethod
    def _from_memcpy_arg(cls, arg: BaseDeviceAllocation | int) -> BaseDeviceAllocation: ...

    @classmethod
    @abstractmethod
    def _from_kernel_arg(cls, arg: BaseDeviceAllocation | numpy.uintp) -> BaseDeviceAllocation: ...

    @abstractmethod
    def __init__(
        self, idx: int, address: int, offset: int, size: int, *, owns_buffer: bool = False
    ): ...

    @abstractmethod
    def _set(self, arr: numpy.ndarray[Any, numpy.dtype[Any]] | BaseDeviceAllocation) -> None: ...

    @abstractmethod
    def _get(self, arr: numpy.ndarray[Any, numpy.dtype[Any]]) -> None: ...

    @abstractmethod
    def __int__(self) -> int: ...


@lru_cache
def make_device_allocation_class(backend: MockPyCUDA) -> type[BaseDeviceAllocation]:  # noqa: C901
    backend_ref = weakref.ref(backend)

    class DeviceAllocation(BaseDeviceAllocation):
        _backend_ref = backend_ref

        @classmethod
        def _allocate(cls, size: int) -> DeviceAllocation:
            idx, address = backend.allocate(size)
            return cls(idx, address, 0, size, owns_buffer=True)

        @classmethod
        def _from_memcpy_arg(cls, arg: BaseDeviceAllocation | int) -> DeviceAllocation:
            # `memcpy` functions in PyCUDA require pointers to be `int`s,
            # and kernels require them to be `numpy.number`s. Go figure.
            assert isinstance(arg, cls | int)
            if isinstance(arg, cls):
                return arg
            arg = cast(int, arg)  # mypy is not smart enough to derive that

            backend = cls._backend_ref()
            assert backend is not None
            # Unfortunately we can't keep track of subregion size in PyCUDA,
            # so for the size we just choose the maximum available.
            idx, offset, size = backend.check_allocation(arg)
            return cls(idx, arg, offset, size, owns_buffer=False)

        @classmethod
        def _from_kernel_arg(cls, arg: BaseDeviceAllocation | numpy.uintp) -> DeviceAllocation:
            assert isinstance(arg, cls | numpy.uintp)
            return cls._from_memcpy_arg(int(arg))

        def __init__(
            self, idx: int, address: int, offset: int, size: int, *, owns_buffer: bool = False
        ):
            backend = self._backend_ref()
            assert backend is not None
            self._context = backend.current_context()
            self._address = address
            self._idx = idx
            self._offset = offset
            self._size = size
            self._owns_buffer = owns_buffer

        def _set(self, arr: numpy.ndarray[Any, numpy.dtype[Any]] | BaseDeviceAllocation) -> None:
            backend = self._backend_ref()
            assert backend is not None
            if isinstance(arr, numpy.ndarray):
                data = arr.tobytes()
            else:
                data = backend.get_allocation_buffer(arr._idx, arr._offset, arr._size)
            assert len(data) <= self._size
            backend.set_allocation_buffer(self._idx, self._offset, data)

        def _get(self, arr: numpy.ndarray[Any, numpy.dtype[Any]]) -> None:
            backend = self._backend_ref()
            assert backend is not None
            data = arr.tobytes()
            assert len(data) <= self._size
            buf = backend.get_allocation_buffer(self._idx, self._offset, len(data))
            buf_as_arr = numpy.frombuffer(buf, arr.dtype).reshape(arr.shape)
            numpy.copyto(arr, buf_as_arr)

        def __int__(self) -> int:
            return self._address

        def __del__(self) -> None:
            # Backend is held alive by the context object we reference.
            if self._owns_buffer:
                backend = self._backend_ref()
                assert backend is not None
                backend.free_allocation(self._idx)

    return DeviceAllocation


class BaseFunction(ABC):
    @abstractmethod
    def __init__(self, source_module: BaseSourceModule, kernel: MockKernel): ...

    @abstractmethod
    def __call__(
        self,
        *args: BaseDeviceAllocation | numpy.generic,
        grid: tuple[int, int, int],
        block: tuple[int, int, int],
        stream: BaseStream,
        shared: int | None = None,
    ) -> None: ...

    @abstractmethod
    def get_attribute(self, attribute: FunctionAttribute) -> tuple[int, ...]: ...


@lru_cache
def make_function_class(backend: MockPyCUDA) -> type[BaseFunction]:
    backend_ref = weakref.ref(backend)

    class Function(BaseFunction):
        _backend_ref = backend_ref

        def __init__(self, source_module: BaseSourceModule, kernel: MockKernel):
            self._kernel = kernel
            self._source_module = source_module

        def __call__(
            self,
            *args: BaseDeviceAllocation | numpy.generic,
            grid: tuple[int, int, int],
            block: tuple[int, int, int],
            stream: BaseStream,
            # This is only needed to mimic PyCUDA API
            shared: int | None = None,  # noqa: ARG002
        ) -> None:
            backend = self._backend_ref()
            assert backend is not None

            current_context = backend.current_context()

            assert self._source_module._context == current_context

            assert isinstance(stream, backend.pycuda_driver.Stream)
            assert stream._context == current_context

            for arg, param in zip(args, self._kernel.parameters, strict=True):
                if param is None:
                    # Checks validity on creation
                    assert isinstance(arg, numpy.uintp)
                    buf = backend.pycuda_driver.DeviceAllocation._from_kernel_arg(arg)
                    assert buf._context == current_context
                elif isinstance(arg, numpy.generic):
                    assert arg.dtype == param
                else:
                    raise TypeError(f"Incorrect argument type: {type(arg)}")

            device = current_context.get_device()
            max_grid = [device.max_grid_dim_x, device.max_grid_dim_y, device.max_grid_dim_z]
            max_block = [device.max_block_dim_x, device.max_block_dim_y, device.max_block_dim_z]

            assert isinstance(grid, tuple)
            assert len(grid) == 3  # noqa: PLR2004
            assert all(isinstance(x, int) for x in grid)
            assert all(g <= max_g for g, max_g in zip(grid, max_grid, strict=True))

            assert isinstance(block, tuple)
            assert len(block) == 3  # noqa: PLR2004
            assert all(isinstance(x, int) for x in block)
            assert all(b <= max_b for b, max_b in zip(block, max_block, strict=True))

        def get_attribute(self, attribute: FunctionAttribute) -> tuple[int, ...]:
            if attribute == FunctionAttribute.MAX_THREADS_PER_BLOCK:
                device_idx = self._source_module._context._device_idx
                return self._kernel.max_total_local_sizes[device_idx]

            # We don't really use any other attributes in this package,
            # and this class is only used internally.
            raise NotImplementedError(f"Unknown attribute: {attribute}")  # pragma: no cover

    return Function


class MockPyCUDATools:
    class DeviceData:
        def __init__(self, device: BaseDevice):
            pass

        def align_words(self, word_size: int) -> int:
            return {4: 16, 8: 16, 16: 8}[word_size]



import pytest

import grunnur
from grunnur.api import all_api_ids, cuda_api_id, opencl_api_id


def test_import_cuda_api(mock_backend_pycuda):
    mock_backend_pycuda.add_devices(["Device1", "Device2"])
    from grunnur import cuda_api

    assert cuda_api.id == cuda_api_id()


def test_import_opencl_api(mock_backend_pyopencl):
    mock_backend_pyopencl.add_devices(["Device1", "Device2"])
    from grunnur import opencl_api

    assert opencl_api.id == opencl_api_id()


def test_import_any_api(mock_backend_factory):
    mock_backend_factory.mock_pycuda().add_devices(["Device1", "Device2"])
    mock_backend_factory.mock_pyopencl().add_devices(["Device1", "Device2"])
    from grunnur import any_api

    assert any_api.id == opencl_api_id() or any_api.id == cuda_api_id()


# `mock_backend_factory` disables all existing backends by default
@pytest.mark.usefixtures("mock_backend_factory")
def test_any_api_none_available():
    with pytest.raises(ImportError):
        from grunnur import any_api


def test_grunnur_import_error():
    # Checks the error branch in `grunnur.__getattr__()`
    with pytest.raises(ImportError):
        from grunnur import non_existent



import itertools

import numpy

from grunnur import Array, Program, Queue
from grunnur.utils import min_blocks, prod
from grunnur.vsize import VirtualSizes


class VirtualSizesTest:
    def __init__(self, global_size, local_size, max_num_groups, max_local_sizes):
        self.global_size = global_size
        self.local_size = local_size
        if local_size is not None:
            self.grid_size = tuple(
                min_blocks(gs, ls) for gs, ls in zip(global_size, local_size, strict=True)
            )

        self.max_num_groups = max_num_groups
        self.max_local_sizes = max_local_sizes
        self.max_total_local_size = prod(max_local_sizes)

    def __str__(self):
        return (
            f"{self.global_size}-{self.local_size}-"
            f"limited-by-{self.max_num_groups}-{self.max_local_sizes}"
        )


_VSTESTS = [
    # This set of parameters ensures that all three skip functions will be nontrivial:
    # - skip_local_threads(), because the local size cannot be trivially fit into max_local_sizes
    #   (that is, one has to construct a flat index from real local ids, and then decompose it
    #   into virtual local ids);
    # - skip_groups(), for the same reason, but applied to groups and max_num_groups;
    # - skip_global_threads(), because global size elements are not multiples
    #   of corresponding local size elements.
    VirtualSizesTest(
        global_size=(51, 201), local_size=(5, 10), max_num_groups=(16, 16), max_local_sizes=(8, 8)
    ),
    # Same thing, but with unspecified local size
    VirtualSizesTest(
        global_size=(51, 201), local_size=None, max_num_groups=(16, 16), max_local_sizes=(8, 8)
    ),
]


def pytest_generate_tests(metafunc):
    if "vstest" in metafunc.fixturenames:
        metafunc.parametrize("vstest", _VSTESTS, ids=[str(x) for x in _VSTESTS])


class ReferenceIds:
    def __init__(self, global_size, local_size):
        self.global_size = global_size
        if local_size is not None:
            self.local_size = local_size
            self.grid_size = tuple(
                min_blocks(gs, ls) for gs, ls in zip(global_size, local_size, strict=True)
            )

    def _tile_pattern(self, pattern, axis, full_shape):
        pattern_shape = [x if i == axis else 1 for i, x in enumerate(full_shape)]
        pattern = pattern.reshape(*pattern_shape)

        tile_shape = [x if i != axis else 1 for i, x in enumerate(full_shape)]
        pattern = numpy.tile(pattern, tile_shape)

        return pattern.astype(numpy.int32)

    def predict_local_ids(self, dim):
        global_len = self.global_size[dim]
        local_len = self.local_size[dim]
        repetitions = min_blocks(global_len, local_len)

        pattern = numpy.tile(numpy.arange(local_len), repetitions)[:global_len]
        return self._tile_pattern(pattern, dim, self.global_size)

    def predict_group_ids(self, dim):
        global_len = self.global_size[dim]
        local_len = self.local_size[dim]
        repetitions = min_blocks(global_len, local_len)

        pattern = numpy.repeat(numpy.arange(repetitions), local_len)[:global_len]
        return self._tile_pattern(pattern, dim, self.global_size)

    def predict_global_ids(self, dim):
        global_len = self.global_size[dim]

        pattern = numpy.arange(global_len)
        return self._tile_pattern(pattern, dim, self.global_size)


def test_ids(context, vstest):
    """Test that virtual IDs are correct for each thread."""
    ref = ReferenceIds(vstest.global_size, vstest.local_size)

    vs = VirtualSizes(
        max_total_local_size=vstest.max_total_local_size,
        max_local_sizes=vstest.max_local_sizes,
        max_num_groups=vstest.max_num_groups,
        local_size_multiple=2,
        virtual_global_size=vstest.global_size,
        virtual_local_size=vstest.local_size,
    )

    program = Program(
        [context.device],
        """
        KERNEL void get_ids(
            GLOBAL_MEM int *local_ids,
            GLOBAL_MEM int *group_ids,
            GLOBAL_MEM int *global_ids,
            int vdim)
        {
            if (${static.skip}()) return;
            const VSIZE_T i = ${static.global_flat_id}();
            local_ids[i] = ${static.local_id}(vdim);
            group_ids[i] = ${static.group_id}(vdim);
            global_ids[i] = ${static.global_id}(vdim);
        }
        """,
        render_globals=dict(static=vs.vsize_modules),
    )

    get_ids = program.kernel.get_ids

    queue = Queue(context.device)
    local_ids = Array.empty(context.device, ref.global_size, numpy.int32)
    group_ids = Array.empty(context.device, ref.global_size, numpy.int32)
    global_ids = Array.empty(context.device, ref.global_size, numpy.int32)

    for vdim in range(len(vstest.global_size)):
        get_ids(
            queue,
            vs.real_global_size,
            vs.real_local_size,
            local_ids,
            group_ids,
            global_ids,
            numpy.int32(vdim),
        )

        assert (global_ids.get(queue) == ref.predict_global_ids(vdim)).all()
        if vstest.local_size is not None:
            assert (local_ids.get(queue) == ref.predict_local_ids(vdim)).all()
            assert (group_ids.get(queue) == ref.predict_group_ids(vdim)).all()


def test_sizes(context, vstest):
    """Test that virtual sizes are correct."""
    vs = VirtualSizes(
        max_total_local_size=vstest.max_total_local_size,
        max_local_sizes=vstest.max_local_sizes,
        max_num_groups=vstest.max_num_groups,
        local_size_multiple=2,
        virtual_global_size=vstest.global_size,
        virtual_local_size=vstest.local_size,
    )

    vdims = len(vstest.global_size)

    program = Program(
        [context.device],
        """
        KERNEL void get_sizes(GLOBAL_MEM int *sizes)
        {
            if (${static.global_id}(0) > 0) return;

            for (int i = 0; i < ${vdims}; i++)
            {
                sizes[i] = ${static.local_size}(i);
                sizes[i + ${vdims}] = ${static.num_groups}(i);
                sizes[i + ${vdims * 2}] = ${static.global_size}(i);
            }
            sizes[${vdims * 3}] = ${static.global_flat_size}();
        }
        """,
        render_globals=dict(vdims=vdims, static=vs.vsize_modules),
    )

    get_sizes = program.kernel.get_sizes

    queue = Queue(context.device)
    sizes = Array.empty(context.device, [vdims * 3 + 1], numpy.int32)
    get_sizes(queue, vs.real_global_size, vs.real_local_size, sizes)

    sizes = sizes.get(queue)
    local_sizes = sizes[0:vdims]
    grid_sizes = sizes[vdims : vdims * 2]
    global_sizes = sizes[vdims * 2 : vdims * 3]
    flat_size = sizes[vdims * 3]

    global_sizes_ref = numpy.array(vstest.global_size)
    assert (global_sizes == global_sizes_ref).all()
    assert flat_size == prod(vstest.global_size)

    if vstest.local_size is not None:
        grid_sizes_ref = numpy.array(vstest.grid_size)
        assert (grid_sizes == grid_sizes_ref).all()
        local_sizes_ref = numpy.array(vstest.local_size)
        assert (local_sizes == local_sizes_ref).all()



"""
A class for a list that keeps itself sorted.
Allows for fast element-based insertions/removals and element finding.
"""

import sys
from bisect import bisect_left, bisect_right
from collections.abc import Callable, Iterable, Iterator, Sequence
from typing import Any, Generic, TypeVar, overload

_T = TypeVar("_T")


class SortedList(Sequence[_T]):
    def __init__(self, iterable: Iterable[_T], key: Callable[[_T], int]):
        decorated = sorted((key(item), item) for item in iterable)
        self._keys = [k for k, item in decorated]
        self._items = [item for k, item in decorated]
        self._key = key

    def __len__(self) -> int:
        return len(self._items)

    @overload
    def __getitem__(self, idx: int) -> _T: ...

    @overload
    def __getitem__(self, idx: slice) -> Sequence[_T]: ...

    def __getitem__(self, idx: int | slice) -> _T | Sequence[_T]:
        return self._items[idx]

    def __iter__(self) -> Iterator[_T]:
        return iter(self._items)

    def index(self, value: _T, start: int = 0, stop: int = sys.maxsize) -> int:
        """
        Finds the position of an item.
        Raises ``ValueError`` if not found.
        """
        k = self._key(value)
        i = bisect_left(self._keys, k, lo=max(start, 0), hi=min(stop, len(self._keys)))
        j = bisect_right(self._keys, k, lo=max(start, 0), hi=min(stop, len(self._keys)))
        return self._items[i:j].index(value) + i

    def insert(self, value: _T) -> None:
        """
        Inserts a new item.
        If equal keys are found, inserts on the left.
        """
        k = self._key(value)
        i = bisect_left(self._keys, k)
        self._keys.insert(i, k)
        self._items.insert(i, value)

    def remove(self, value: _T) -> None:
        """
        Removes the first occurence of an item.
        Raises ``ValueError`` if not found.
        """
        i = self.index(value)
        del self._keys[i]
        del self._items[i]

    def argfind_ge(self, key: int) -> int:
        """
        Returns the index of the first item with ``key >= k``.
        Raises ``ValueError`` if not found.
        """
        i = bisect_left(self._keys, key)
        if i != len(self):
            return i
        raise ValueError(f"No item found with key at or above: {key}")



import itertools
import re
from functools import partial
from warnings import catch_warnings, filterwarnings

import numpy
import pytest

from grunnur import Array, Program, Queue, dtypes, functions
from grunnur.modules import render_with_modules
from grunnur.template import RenderError
from grunnur.utils import prod
from utils import get_test_array


def get_func_kernel(func_module, out_dtype, in_dtypes):
    src = """
    <%
        argnames = ["a" + str(i + 1) for i in range(len(in_dtypes))]
        in_ctypes = list(map(dtypes.ctype, in_dtypes))
        out_ctype = dtypes.ctype(out_dtype)
    %>
    KERNEL void test(
        GLOBAL_MEM ${out_ctype} *dest
        %for arg, ctype in zip(argnames, in_ctypes, strict=True):
        , GLOBAL_MEM ${ctype} *${arg}
        %endfor
        )
    {
        const SIZE_T i = get_global_id(0);
        %for arg, ctype in zip(argnames, in_ctypes, strict=True):
        ${ctype} ${arg}_load = ${arg}[i];
        %endfor

        dest[i] = ${func}(${", ".join([arg + "_load" for arg in argnames])});
    }
    """

    return render_with_modules(
        src,
        render_globals=dict(
            dtypes=dtypes, in_dtypes=in_dtypes, out_dtype=out_dtype, func=func_module
        ),
    )


def generate_dtypes(out_code, in_codes):
    test_dtype = lambda idx: dict(i=numpy.int32, f=numpy.float32, c=numpy.complex64)[idx]
    in_dtypes = list(map(test_dtype, in_codes))
    out_dtype = dtypes.result_type(*in_dtypes) if out_code == "auto" else test_dtype(out_code)

    if not any(map(dtypes.is_double, in_dtypes)):
        # numpy thinks that int32 * float32 == float64,
        # but we still need to run this test on older videocards
        if dtypes.is_complex(out_dtype):
            out_dtype = numpy.complex64
        elif dtypes.is_real(out_dtype):
            out_dtype = numpy.float32

    return out_dtype, in_dtypes


def check_func(
    context,
    func_module,
    reference_func,
    out_dtype,
    in_dtypes,
    *,
    atol=1e-8,
    rtol=1e-5,
    is_mocked=False,
):
    test_size = 256

    full_src = get_func_kernel(func_module, out_dtype, in_dtypes)

    # Can't test anything else if we don't have a real context
    if is_mocked:
        return

    program = Program([context.device], full_src)
    test = program.kernel.test

    queue = Queue(context.device)

    arrays = [get_test_array(test_size, dt, no_zeros=True, high=8) for dt in in_dtypes]
    arrays_dev = [Array.from_host(queue, array) for array in arrays]
    dest_dev = Array.empty(context.device, [test_size], out_dtype)

    test(queue, [test_size], None, dest_dev, *arrays_dev)

    assert numpy.allclose(
        dest_dev.get(queue), reference_func(*arrays).astype(out_dtype), atol=atol, rtol=rtol
    )


# exp()


_test_exp_parameters = pytest.mark.parametrize(("out_code", "in_codes"), [("f", "f"), ("c", "c")])


def _test_exp(context, out_code, in_codes, is_mocked):
    out_dtype, in_dtypes = generate_dtypes(out_code, in_codes)
    check_func(
        context, functions.exp(in_dtypes[0]), numpy.exp, out_dtype, in_dtypes, is_mocked=is_mocked
    )


@_test_exp_parameters
def test_exp(context, out_code, in_codes):
    _test_exp(context, out_code, in_codes, is_mocked=False)


@_test_exp_parameters
def test_exp_mocked(mock_context, out_code, in_codes):
    _test_exp(mock_context, out_code, in_codes, is_mocked=True)


def test_exp_of_integer():
    with pytest.raises(
        ValueError, match=re.escape("exp() of <class 'numpy.int32'> is not supported")
    ):
        functions.exp(numpy.int32)


# pow()


_test_pow_parameters = pytest.mark.parametrize(
    ("out_code", "in_codes"), [("f", "fi"), ("c", "ci"), ("f", "ff"), ("c", "cf"), ("i", "ii")]
)


def _test_pow(context, out_code, in_codes, is_mocked):
    out_dtype, in_dtypes = generate_dtypes(out_code, in_codes)
    func = functions.pow(in_dtypes[0], exponent_dtype=in_dtypes[1], out_dtype=out_dtype)
    check_func(context, func, numpy.power, out_dtype, in_dtypes, is_mocked=is_mocked)


@_test_pow_parameters
def test_pow(context, out_code, in_codes):
    _test_pow(context, out_code, in_codes, is_mocked=False)


@_test_pow_parameters
def test_pow_mocked(mock_context, out_code, in_codes):
    _test_pow(mock_context, out_code, in_codes, is_mocked=True)


def _test_pow_defaults(context, is_mocked):
    func = functions.pow(numpy.float32)  # check that exponent and output default to the base dtype
    check_func(
        context,
        func,
        numpy.power,
        numpy.float32,
        [numpy.float32, numpy.float32],
        is_mocked=is_mocked,
    )


def test_pow_defaults(context):
    _test_pow_defaults(context, is_mocked=False)


def test_pow_defaults_mocked(mock_context):
    _test_pow_defaults(mock_context, is_mocked=True)


def _test_pow_cast_output(context, is_mocked):
    func = functions.pow(numpy.int32, exponent_dtype=numpy.int32, out_dtype=numpy.int64)
    check_func(
        context, func, numpy.power, numpy.int64, [numpy.int32, numpy.int32], is_mocked=is_mocked
    )


def test_pow_cast_output(context):
    _test_pow_cast_output(context, is_mocked=False)


def test_pow_cast_output_mocked(mock_context):
    _test_pow_cast_output(mock_context, is_mocked=True)


def test_pow_complex_exponent():
    with pytest.raises(
        ValueError, match=re.escape("pow() with a complex exponent is not supported")
    ):
        functions.pow(numpy.float32, exponent_dtype=numpy.complex64)


def test_pow_int_to_float():
    with pytest.raises(
        ValueError, match=re.escape("pow(integer, float): integer is not supported")
    ):
        functions.pow(numpy.int32, exponent_dtype=numpy.float32, out_dtype=numpy.int32)


@pytest.mark.parametrize(("out_code", "in_codes"), [("c", "cf"), ("f", "ff")])
def test_pow_zero_base(context, out_code, in_codes):
    """Specific tests for 0^0 and 0^x."""
    test_size = 256

    out_dtype, in_dtypes = generate_dtypes(out_code, in_codes)
    func_module = functions.pow(in_dtypes[0], exponent_dtype=in_dtypes[1], out_dtype=out_dtype)
    full_src = get_func_kernel(func_module, out_dtype, in_dtypes)
    program = Program([context.device], full_src)
    test = program.kernel.test

    queue = Queue(context.device)
    bases = Array.from_host(queue, numpy.zeros(test_size, in_dtypes[0]))

    # zero exponents
    exponents = Array.from_host(queue, numpy.zeros(test_size, in_dtypes[1]))
    dest_dev = Array.empty(context.device, [test_size], out_dtype)
    test(queue, [test_size], None, dest_dev, bases, exponents)
    assert numpy.allclose(dest_dev.get(queue), numpy.ones(test_size, in_dtypes[0]))

    # non-zero exponents
    exponents = Array.from_host(queue, numpy.ones(test_size, in_dtypes[1]))
    dest_dev = Array.empty(context.device, [test_size], out_dtype)
    test(queue, [test_size], None, dest_dev, bases, exponents)
    assert numpy.allclose(dest_dev.get(queue), numpy.zeros(test_size, in_dtypes[0]))


# polar_unit()


_test_polar_unit_parameters = pytest.mark.parametrize(("out_code", "in_codes"), [("c", "f")])


def _test_polar_unit(context, out_code, in_codes, is_mocked):
    out_dtype, in_dtypes = generate_dtypes(out_code, in_codes)
    check_func(
        context,
        functions.polar_unit(in_dtypes[0]),
        lambda theta: numpy.exp(1j * theta),
        out_dtype,
        in_dtypes,
        is_mocked=is_mocked,
    )


@_test_polar_unit_parameters
def test_polar_unit(context, out_code, in_codes):
    _test_polar_unit(context, out_code, in_codes, is_mocked=False)


@_test_polar_unit_parameters
def test_polar_unit_mocked(mock_context, out_code, in_codes):
    _test_polar_unit(mock_context, out_code, in_codes, is_mocked=True)


def test_polar_unit_of_complex():
    with pytest.raises(
        ValueError, match=re.escape("polar_unit() can only be applied to real dtypes")
    ):
        functions.polar_unit(numpy.complex64)


# polar()


_test_polar_parameters = pytest.mark.parametrize(("out_code", "in_codes"), [("c", "ff")])


def _test_polar(context, out_code, in_codes, is_mocked):
    out_dtype, in_dtypes = generate_dtypes(out_code, in_codes)
    check_func(
        context,
        functions.polar(in_dtypes[0]),
        lambda rho, theta: rho * numpy.exp(1j * theta),
        out_dtype,
        in_dtypes,
        is_mocked=is_mocked,
    )


@_test_polar_parameters
def test_polar(context, out_code, in_codes):
    _test_polar(context, out_code, in_codes, is_mocked=False)


@_test_polar_parameters
def test_polar_mocked(mock_context, out_code, in_codes):
    _test_polar(mock_context, out_code, in_codes, is_mocked=True)


def test_polar_of_complex():
    with pytest.raises(
        ValueError, match=re.escape("polar() of <class 'numpy.complex64'> is not supported")
    ):
        functions.polar(numpy.complex64)


# norm()


_test_norm_parameters = pytest.mark.parametrize(
    ("out_code", "in_codes"), [("f", "c"), ("f", "f"), ("i", "i")]
)


def _test_norm(context, out_code, in_codes, is_mocked):
    out_dtype, in_dtypes = generate_dtypes(out_code, in_codes)
    check_func(
        context,
        functions.norm(in_dtypes[0]),
        lambda x: numpy.abs(x) ** 2,
        out_dtype,
        in_dtypes,
        is_mocked=is_mocked,
    )


@_test_norm_parameters
def test_norm(context, out_code, in_codes):
    _test_norm(context, out_code, in_codes, is_mocked=False)


@_test_norm_parameters
def test_norm_mocked(mock_context, out_code, in_codes):
    _test_norm(mock_context, out_code, in_codes, is_mocked=True)


# conj()


_test_conj_parameters = pytest.mark.parametrize(("out_code", "in_codes"), [("c", "c"), ("f", "f")])


def _test_conj(context, out_code, in_codes, is_mocked):
    out_dtype, in_dtypes = generate_dtypes(out_code, in_codes)
    check_func(
        context, functions.conj(in_dtypes[0]), numpy.conj, out_dtype, in_dtypes, is_mocked=is_mocked
    )


@_test_conj_parameters
def test_conj(context, out_code, in_codes):
    _test_conj(context, out_code, in_codes, is_mocked=False)


@_test_conj_parameters
def test_conj_mocked(mock_context, out_code, in_codes):
    _test_conj(mock_context, out_code, in_codes, is_mocked=True)


# cast()


_test_cast_parameters = pytest.mark.parametrize(
    ("out_code", "in_codes"), [("c", "f"), ("f", "f"), ("c", "c")]
)


def _test_cast(context, out_code, in_codes, is_mocked):
    out_dtype, in_dtypes = generate_dtypes(out_code, in_codes)
    check_func(
        context,
        functions.cast(in_dtypes[0], out_dtype),
        partial(numpy.asarray, dtype=out_dtype),
        out_dtype,
        in_dtypes,
        is_mocked=is_mocked,
    )


@_test_cast_parameters
def test_cast(context, out_code, in_codes):
    _test_cast(context, out_code, in_codes, is_mocked=False)


@_test_cast_parameters
def test_cast_mocked(mock_context, out_code, in_codes):
    _test_cast(mock_context, out_code, in_codes, is_mocked=True)


def test_cast_complex_to_real(context):
    out_dtype = numpy.float32
    in_dtypes = [numpy.complex64]
    message = re.escape(
        "cast from <class 'numpy.complex64'> " "to <class 'numpy.float32'> is not supported"
    )
    with pytest.raises(ValueError, match=message):
        check_func(
            context,
            functions.cast(in_dtypes[0], out_dtype),
            partial(numpy.asarray, dtype=out_dtype),
            out_dtype,
            in_dtypes,
        )


# div()


_test_div_parameters = pytest.mark.parametrize(
    ("out_code", "in_codes"), [("f", "ff"), ("c", "cc"), ("c", "cf"), ("c", "fc"), ("f", "if")]
)


def _test_div(context, out_code, in_codes, is_mocked):
    out_dtype, in_dtypes = generate_dtypes(out_code, in_codes)
    check_func(
        context,
        functions.div(*in_dtypes, out_dtype=out_dtype),
        lambda x, y: numpy.asarray(x / y, out_dtype),
        out_dtype,
        in_dtypes,
        is_mocked=is_mocked,
    )


@_test_div_parameters
def test_div(context, out_code, in_codes):
    _test_div(context, out_code, in_codes, is_mocked=False)


@_test_div_parameters
def test_div_mocked(mock_context, out_code, in_codes):
    _test_div(mock_context, out_code, in_codes, is_mocked=True)


# add()


def make_reference_add(out_dtype):
    def reference_add(*args):
        res = sum(args)
        if not dtypes.is_complex(out_dtype) and dtypes.is_complex(res.dtype):
            res = res.real
        return res.astype(out_dtype)

    return reference_add


_test_add_parameters = pytest.mark.parametrize("in_codes", ["ff", "cc", "cf", "fc", "ifccfi"])


def _test_add(context, in_codes, is_mocked):
    """Checks multi-argument add() with a variety of data types."""
    out_dtype, in_dtypes = generate_dtypes("auto", in_codes)
    reference_add = make_reference_add(out_dtype)
    add = functions.add(*in_dtypes, out_dtype=out_dtype)
    check_func(context, add, reference_add, out_dtype, in_dtypes, is_mocked=is_mocked)


@_test_add_parameters
def test_add(context, in_codes):
    _test_add(context, in_codes, is_mocked=False)


@_test_add_parameters
def test_add_mocked(mock_context, in_codes):
    _test_add(mock_context, in_codes, is_mocked=True)


_test_add_cast_parameters = pytest.mark.parametrize(
    ("out_code", "in_codes"), [("c", "ff"), ("c", "cc"), ("f", "ff"), ("f", "cc")]
)


def _test_add_cast(context, out_code, in_codes, is_mocked):
    """Check that add() casts the result correctly."""
    out_dtype, in_dtypes = generate_dtypes(out_code, in_codes)
    reference_add = make_reference_add(out_dtype)

    # Temporarily catching imaginary part truncation warnings
    with catch_warnings():
        filterwarnings("ignore", "", numpy.exceptions.ComplexWarning)
        add = functions.add(*in_dtypes, out_dtype=out_dtype)

    check_func(context, add, reference_add, out_dtype, in_dtypes, is_mocked=is_mocked)


@_test_add_cast_parameters
def test_add_cast(context, out_code, in_codes):
    _test_add_cast(context, out_code, in_codes, is_mocked=False)


@_test_add_cast_parameters
def test_add_cast_mocked(mock_context, out_code, in_codes):
    _test_add_cast(mock_context, out_code, in_codes, is_mocked=True)


# mul()


def make_reference_mul(out_dtype):
    def reference_mul(*args):
        res = prod(args)
        if not dtypes.is_complex(out_dtype) and dtypes.is_complex(res.dtype):
            res = res.real
        return res.astype(out_dtype)

    return reference_mul


_test_mul_parameters = pytest.mark.parametrize("in_codes", ["ff", "cc", "cf", "fc", "ifccfi"])


def _test_mul(context, in_codes, is_mocked):
    """Checks multi-argument mul() with a variety of data types."""
    out_dtype, in_dtypes = generate_dtypes("auto", in_codes)
    reference_mul = make_reference_mul(out_dtype)
    mul = functions.mul(*in_dtypes, out_dtype=out_dtype)
    check_func(context, mul, reference_mul, out_dtype, in_dtypes, is_mocked=is_mocked)


@_test_mul_parameters
def test_mul(context, in_codes):
    _test_mul(context, in_codes, is_mocked=False)


@_test_mul_parameters
def test_mul_mocked(mock_context, in_codes):
    _test_mul(mock_context, in_codes, is_mocked=True)


_test_mul_cast_parameters = pytest.mark.parametrize(
    ("out_code", "in_codes"), [("c", "ff"), ("c", "cc"), ("f", "ff"), ("f", "cc")]
)


def _test_mul_cast(context, out_code, in_codes, is_mocked):
    """Check that mul() casts the result correctly."""
    out_dtype, in_dtypes = generate_dtypes(out_code, in_codes)
    reference_mul = make_reference_mul(out_dtype)

    # Temporarily catching imaginary part truncation warnings
    with catch_warnings():
        filterwarnings("ignore", "", numpy.exceptions.ComplexWarning)
        mul = functions.mul(*in_dtypes, out_dtype=out_dtype)

    # Relax tolerance a little - in single precision the difference may sometimes go to 1e-5
    check_func(context, mul, reference_mul, out_dtype, in_dtypes, is_mocked=is_mocked, rtol=1e-4)


@_test_mul_cast_parameters
def test_mul_cast(context, out_code, in_codes):
    _test_mul_cast(context, out_code, in_codes, is_mocked=False)


@_test_mul_cast_parameters
def test_mul_cast_mocked(mock_context, out_code, in_codes):
    _test_mul_cast(mock_context, out_code, in_codes, is_mocked=True)



import numpy

from grunnur import Array, Program, Queue, dtypes


def check_offsets_on_device(queue, dtype):
    flat_dtype = dtypes.flatten_dtype(dtype)

    program = Program(
        [queue.device],
        """
        #define my_offsetof(type, field) ((size_t)(&((type *)0)->field))

        KERNEL void test(GLOBAL_MEM int *dest)
        {
            ${struct} temp;

            %for i, field_info in enumerate(flat_dtype):
            dest[${i}] = (size_t)&(temp${field_info.c_path}) - (size_t)&temp;
            %endfor
            dest[${len(flat_dtype)}] = sizeof(${struct});
        }
        """,
        render_globals=dict(struct=dtypes.ctype(dtype), dtype=dtype, flat_dtype=flat_dtype),
    )

    offsets_dev = Array.empty(queue.device, (len(flat_dtype) + 1,), numpy.int32)
    test = program.kernel.test
    test(queue, [1], None, offsets_dev)
    offsets = offsets_dev.get(queue)
    offsets = [int(offset) for offset in offsets]

    for field_info, device_offset in zip(flat_dtype, offsets[:-1], strict=True):
        message = (
            f"offset for {field_info.c_path} is different: {field_info.offset} in numpy, "
            f"{device_offset} on device"
        )
        assert field_info.offset == device_offset, message

    assert offsets[-1] == dtype.itemsize


def test_offsets_simple(context):
    queue = Queue(context.device)
    dtype = numpy.dtype(
        dict(
            names=["val1", "pad"],
            formats=[numpy.int32, numpy.int8],
            offsets=[0, 4],
            itemsize=8,
            aligned=True,
        )
    )
    check_offsets_on_device(queue, dtype)


def test_offsets_nested(context):
    queue = Queue(context.device)
    dtype_nested = numpy.dtype(
        dict(
            names=["val1", "pad"],
            formats=[numpy.int32, numpy.int8],
            offsets=[0, 4],
            itemsize=8,
            aligned=True,
        )
    )
    dtype = numpy.dtype(
        dict(
            names=["val1", "val2", "nested"],
            formats=[numpy.int32, numpy.int16, dtype_nested],
            offsets=[0, 4, 8],
            itemsize=32,
            aligned=True,
        )
    )
    check_offsets_on_device(queue, dtype)


def test_offsets_arrays(context):
    queue = Queue(context.device)
    dtype = numpy.dtype(
        dict(
            names=["val1", "val2"],
            formats=[(numpy.int8, 3), (numpy.int8, 2)],
            offsets=[0, 8],
            itemsize=16,
            aligned=True,
        )
    )
    check_offsets_on_device(queue, dtype)


def test_offsets_nested_arrays(context):
    queue = Queue(context.device)
    dtype_nested = numpy.dtype(dict(names=["val1", "pad"], formats=[numpy.int8, numpy.int8]))
    dtype = numpy.dtype(
        dict(
            names=["pad", "struct_arr", "regular_arr"],
            formats=[numpy.int32, numpy.dtype((dtype_nested, 2)), numpy.dtype((numpy.int16, 3))],
        )
    )
    dtype_aligned = dtypes.align(dtype)
    check_offsets_on_device(queue, dtype_aligned)


def test_offsets_custom_itemsize(context):
    queue = Queue(context.device)
    dtype = numpy.dtype(
        dict(
            names=["x", "y", "z"],
            formats=[numpy.int8, numpy.int16, numpy.int32],
            offsets=[0, 4, 16],
            itemsize=32,
            aligned=True,
        )
    )
    check_offsets_on_device(queue, dtype)


def test_offsets_custom_itemsize_nested(context):
    queue = Queue(context.device)
    dtype_nested = numpy.dtype(
        dict(
            names=["val1", "pad"],
            formats=[numpy.int32, numpy.int8],
            offsets=[0, 4],
            itemsize=16,
            aligned=True,
        )
    )

    dtype = numpy.dtype(
        dict(
            names=["val1", "val2", "nested"],
            formats=[numpy.int32, numpy.int16, dtype_nested],
            offsets=[0, 4, 16],
            itemsize=64,
            aligned=True,
        )
    )

    check_offsets_on_device(queue, dtype)



import pytest

from grunnur.utils import min_blocks, prod
from grunnur.vsize import (
    PrimeFactors,
    ShapeGroups,
    VirtualSizeError,
    VirtualSizes,
    factorize,
    find_bounding_shape,
    find_local_size_decomposition,
    get_decompositions,
    group_dimensions,
)


def test_factorize():
    assert factorize(2**10) == [2] * 10
    assert factorize(2**3 * 3**2 * 7 * 13 * 251) == [2, 2, 2, 3, 3, 7, 13, 251]


def test_prime_factors():
    val = 2**3 * 3**2 * 7 * 13 * 251
    pf = PrimeFactors.decompose(val)

    assert pf.get_value() == val
    bases, exponents = pf.get_arrays()
    assert (list(bases), list(exponents)) == ([2, 3, 7, 13, 251], [3, 2, 1, 1, 1])

    val2 = 2**2 * 3 * 251
    pf2 = PrimeFactors.decompose(val2)

    assert pf.div_by(pf2) == PrimeFactors.decompose(val // val2)


def test_get_decompositions():
    val = 2**3 * 3**2 * 7 * 13
    parts = 3
    decomps = list(get_decompositions(val, parts))

    assert len({tuple(d) for d in decomps}) == len(decomps)
    assert all(prod(decomp) == val for decomp in decomps)
    assert all(len(decomp) == parts for decomp in decomps)


def test_find_local_size_decomposition():
    # Main pathway: the algorithm can find a local size
    # with less than `threshold` ratio of empty threads
    global_size = (128, 200, 300)
    flat_local_size = 1024
    threshold = 0.05
    local_size = find_local_size_decomposition(global_size, flat_local_size, threshold=threshold)

    full_global_size = [
        min_blocks(gs, ls) * ls for gs, ls in zip(global_size, local_size, strict=True)
    ]

    assert len(local_size) == len(global_size)
    assert prod(local_size) == flat_local_size
    assert (prod(full_global_size) - prod(global_size)) / prod(global_size) <= threshold

    # Now the algorithm can't find a local size with less than `threshold` empty threads,
    # and just returns the best one
    global_size = [100, 200, 300]
    flat_local_size = 1024
    threshold = 0.05
    local_size = find_local_size_decomposition(global_size, flat_local_size, threshold=threshold)

    assert len(local_size) == len(global_size)
    assert prod(local_size) == flat_local_size

    # Shortcut - if `product(global_size) == flat_local_size`, just return the global size
    global_size = [10, 10, 10]
    local_size = find_local_size_decomposition(global_size, 1000)
    assert local_size == global_size

    with pytest.raises(
        ValueError, match="The given local size does not fit in the given global size"
    ):
        find_local_size_decomposition([10, 10, 10], 1024)


def test_group_dimensions():
    # more virtual dimensions than available dimensions
    vgroups, agroups = group_dimensions((8, 8, 8, 8), (32, 32, 32))
    assert vgroups == [[0], [1, 2, 3]]
    assert agroups == [[0], [1, 2]]

    # less virtual dimensions than available dimensions
    vgroups, agroups = group_dimensions((8, 8), (32, 32, 32))
    assert vgroups == [[0], [1]]
    assert agroups == [[0], [1]]

    # some trivial dimensions (of size 1)
    vgroups, agroups = group_dimensions((8, 1, 1, 8, 1), (32, 32, 32))
    assert vgroups == [[0, 1, 2], [3, 4]]
    assert agroups == [[0], [1]]


def test_find_bounding_shape():
    assert find_bounding_shape(100, (8, 16)) == [8, 13]
    assert find_bounding_shape(120, (11, 11)) == [11, 11]


def test_shape_groups():
    sg = ShapeGroups((8, 8, 8, 8), (32, 32, 32))
    assert sg.real_dims == {0: [0], 1: [1, 2], 2: [1, 2], 3: [1, 2]}
    assert sg.real_strides == {0: [1], 1: [1, 23], 2: [1, 23], 3: [1, 23]}
    assert sg.virtual_strides == {0: 1, 1: 1, 2: 8, 3: 64}
    assert sg.major_vdims == {0: 0, 1: 3, 2: 3, 3: 3}
    assert sg.bounding_shape == [8, 23, 23]
    assert sg.skip_thresholds == [(512, [(1, 1), (2, 23)])]

    # A test to cover the case of all virtual dimensions being 1,
    # to test for a corner case in a search for the major vdim.
    sg = ShapeGroups((1, 1, 1), (32,))
    assert sg.real_dims == {0: [0], 1: [0], 2: [0]}
    assert sg.real_strides == {0: [1], 1: [1], 2: [1]}
    assert sg.virtual_strides == {0: 1, 1: 1, 2: 1}
    assert sg.major_vdims == {0: 0, 1: 0, 2: 0}
    assert sg.bounding_shape == [1]
    assert sg.skip_thresholds == []


def test_virtual_sizes():
    vs = VirtualSizes(
        max_total_local_size=1024,
        max_local_sizes=(1024, 1024, 64),
        max_num_groups=(2**31 - 1, 65536, 65536),
        local_size_multiple=16,
        virtual_global_size=(8, 16, 32, 60),
        virtual_local_size=None,
    )

    assert vs._virtual_local_size == [4, 32, 8, 1]
    assert vs._virtual_global_size == [60, 32, 16, 8]
    assert vs._bounding_global_size == [60, 32, 16, 8]
    assert vs._virtual_grid_size == [15, 1, 2, 8]
    assert vs.real_local_size == (4, 32, 8)
    assert vs.real_global_size == (60, 64, 64)

    # Test the case of small total global size (< max_total_local_size)
    vs = VirtualSizes(
        max_total_local_size=1024,
        max_local_sizes=(1024, 1024, 64),
        max_num_groups=(2**31 - 1, 65536, 65536),
        local_size_multiple=16,
        virtual_global_size=(8, 16),
        virtual_local_size=None,
    )

    assert vs._virtual_local_size == [16, 8]
    assert vs._virtual_global_size == [16, 8]
    assert vs._bounding_global_size == [16, 8]
    assert vs._virtual_grid_size == [1, 1]
    assert vs.real_local_size == (16, 8)
    assert vs.real_global_size == (16, 8)

    # Test the case of specified virtual_local_size
    vs = VirtualSizes(
        max_total_local_size=1024,
        max_local_sizes=(1024, 1024, 64),
        max_num_groups=(2**31 - 1, 65536, 65536),
        local_size_multiple=16,
        virtual_global_size=(123, 345),
        virtual_local_size=(13, 15),
    )

    assert vs._virtual_local_size == [15, 13]
    assert vs._virtual_global_size == [345, 123]
    assert vs._bounding_global_size == [345, 130]
    assert vs._virtual_grid_size == [23, 10]
    assert vs.real_local_size == (15, 13)
    assert vs.real_global_size == (345, 130)


def test_vsize_errors():
    with pytest.raises(
        ValueError, match="Global size and local size must have the same number of dimensions"
    ):
        VirtualSizes(
            max_total_local_size=1024,
            max_local_sizes=(1024, 1024, 64),
            max_num_groups=(2**31 - 1, 65536, 65536),
            local_size_multiple=16,
            # different number of dimensions in global and local sizes
            virtual_global_size=(8, 16, 32, 60),
            virtual_local_size=(4, 4, 4),
        )

    with pytest.raises(VirtualSizeError):
        VirtualSizes(
            max_total_local_size=1024,
            max_local_sizes=(1024, 1024, 64),
            max_num_groups=(2**31 - 1, 65536, 65536),
            local_size_multiple=16,
            virtual_global_size=(8, 16, 32, 60),
            # Requested total local size >= max_total_local_size
            virtual_local_size=(8, 8, 8, 8),
        )

    with pytest.raises(VirtualSizeError):
        VirtualSizes(
            max_total_local_size=1024,
            max_local_sizes=(1024, 1024, 64),
            max_num_groups=(4, 4, 4),
            local_size_multiple=16,
            # resulting number of groups is greater than total defined by `prod(max_num_groups)`
            virtual_global_size=(8, 16, 32, 60),
            virtual_local_size=(4, 4, 4, 4),
        )



from __future__ import annotations

import weakref
from collections.abc import Callable, Mapping, Sequence
from math import log10
from typing import TYPE_CHECKING, Any, Generic, TypeVar, cast

import numpy

from .adapter_base import (
    AdapterCompilationError,
    BufferAdapter,
    KernelAdapter,
    ProgramAdapter,
)
from .api import cuda_api_id
from .array import Array, MultiArray
from .buffer import Buffer
from .context import BoundDevice, BoundMultiDevice, Context
from .device import Device
from .modules import Snippet, render_with_modules
from .queue import MultiQueue, Queue
from .utils import update_dict

if TYPE_CHECKING:  # pragma: no cover
    from .array_metadata import ArrayMetadataLike
    from .template import DefTemplate


class CompilationError(RuntimeError):
    def __init__(self, backend_exception: Exception):
        super().__init__(str(backend_exception))
        self.backend_exception = backend_exception


def _check_set_constant_array(queue: Queue, program_devices: BoundMultiDevice) -> None:
    if queue.device.context != program_devices.context:
        raise ValueError("The provided queue must belong to the same context as this program uses")
    if queue.device not in program_devices:
        raise ValueError(
            f"The program was not compiled for the device this queue uses ({queue.device})"
        )


def _set_constant_array(
    queue: Queue,
    program_adapter: ProgramAdapter,
    name: str,
    arr: Array | Buffer | numpy.ndarray[Any, numpy.dtype[Any]],
) -> None:
    """Uploads a constant array ``arr`` corresponding to the symbol ``name`` to the context."""
    queue_adapter = queue._queue_adapter  # noqa: SLF001

    constant_data: BufferAdapter | numpy.ndarray[Any, numpy.dtype[Any]]

    if isinstance(arr, Array):
        constant_data = arr.data._buffer_adapter  # noqa: SLF001
    elif isinstance(arr, Buffer):
        constant_data = arr._buffer_adapter  # noqa: SLF001
    elif isinstance(arr, numpy.ndarray):
        constant_data = arr
    else:
        raise TypeError(f"Unsupported array type: {type(arr)}")

    program_adapter.set_constant_buffer(queue_adapter, name, constant_data)


class SingleDeviceProgram:
    """A program compiled for a single device."""

    device: BoundDevice

    source: str

    def __init__(
        self,
        device: BoundDevice,
        template_src: str | Callable[..., str] | DefTemplate | Snippet,
        *,
        no_prelude: bool = False,
        fast_math: bool = False,
        render_args: Sequence[Any] = [],
        render_globals: Mapping[str, Any] = {},
        constant_arrays: Mapping[str, ArrayMetadataLike] | None = None,
        keep: bool = False,
        compiler_options: Sequence[str] | None = None,
    ):
        """
        Renders and compiles the given template on a single device.

        :param device:
        :param template_src: see :py:meth:`compile`.
        :param no_prelude: see :py:meth:`compile`.
        :param fast_math: see :py:meth:`compile`.
        :param render_args: see :py:meth:`compile`.
        :param render_globals: see :py:meth:`compile`.
        :param kwds: additional parameters for compilation, see :py:func:`compile`.
        """
        if device.context.api.id != cuda_api_id() and constant_arrays and len(constant_arrays) > 0:
            raise ValueError("Compile-time constant arrays are only supported for CUDA API")

        self.device = device

        render_globals = update_dict(
            render_globals,
            dict(device_params=device.params),
            error_msg="'device_params' is a reserved global name and cannot be used",
        )

        src = render_with_modules(
            template_src, render_args=render_args, render_globals=render_globals
        )

        context_adapter = device.context._context_adapter  # noqa: SLF001

        prelude = "" if no_prelude else context_adapter.render_prelude(fast_math=fast_math)

        try:
            self._sd_program_adapter = context_adapter.compile_single_device(
                device._device_adapter,  # noqa: SLF001
                prelude,
                src,
                fast_math=fast_math,
                constant_arrays=constant_arrays,
                keep=keep,
                compiler_options=compiler_options,
            )
        except AdapterCompilationError as exc:
            print(f"Failed to compile on {device}")  # noqa: T201

            lines = exc.source.split("\n")
            max_num_len = int(log10(len(lines))) + 1
            for line_num, line in enumerate(lines):
                print(str(line_num + 1).rjust(max_num_len) + ": " + line)  # noqa: T201

            raise CompilationError(exc.backend_exception) from exc

        self.source = self._sd_program_adapter.source

    def get_kernel_adapter(self, kernel_name: str) -> KernelAdapter:
        """
        Returns a :py:class:`SingleDeviceKernel` object for a function (CUDA)/kernel (OpenCL)
        with the name ``kernel_name``.
        """
        return cast(KernelAdapter, getattr(self._sd_program_adapter, kernel_name))

    def set_constant_array(
        self,
        queue: Queue,
        name: str,
        arr: Array | Buffer | numpy.ndarray[Any, numpy.dtype[Any]],
    ) -> None:
        """Uploads a constant array ``arr`` corresponding to the symbol ``name`` to the context."""
        _set_constant_array(queue, self._sd_program_adapter, name, arr)


class Program:
    """A compiled program on device(s)."""

    devices: BoundMultiDevice
    """The devices on which this program was compiled."""

    sources: dict[BoundDevice, str]
    """Source files used for each device."""

    kernel: KernelHub
    """
    An object whose attributes are :py:class:`~grunnur.program.Kernel` objects
    with the corresponding names.
    """

    def __init__(
        self,
        devices: Sequence[BoundDevice],
        template_src: str | Callable[..., str] | DefTemplate | Snippet,
        *,
        no_prelude: bool = False,
        fast_math: bool = False,
        render_args: Sequence[Any] = (),
        render_globals: Mapping[str, Any] = {},
        compiler_options: Sequence[str] | None = None,
        keep: bool = False,
        constant_arrays: Mapping[str, ArrayMetadataLike] | None = None,
    ):
        """
        :param devices: a single- or a multi-device object on which to compile this program.
        :param template_src: a string with the source code, or a Mako template source to render.
        :param no_prelude: do not add prelude to the rendered source.
        :param fast_math: compile using fast (but less accurate) math functions.
        :param render_args: a list of positional args to pass to the template.
        :param render_globals: a dictionary of globals to pass to the template.
        :param compiler_options: a list of options to pass to the backend compiler.
        :param keep: keep the intermediate files in a temporary directory.
        :param constant_arrays: (**CUDA only**) a dictionary ``name: (size, dtype)``
            of global constant arrays to be declared in the program.
        """
        sd_programs = {}
        sources = {}

        multi_device = BoundMultiDevice.from_bound_devices(devices)

        for device in multi_device:
            sd_program = SingleDeviceProgram(
                device,
                template_src,
                no_prelude=no_prelude,
                fast_math=fast_math,
                render_args=render_args,
                render_globals=render_globals,
                compiler_options=compiler_options,
                keep=keep,
                constant_arrays=constant_arrays,
            )
            sd_programs[device] = sd_program
            sources[device] = sd_program.source

        self._sd_programs = sd_programs
        self.sources = sources
        self.devices = multi_device

        # TODO: create dynamically, in case someone wants to hold a reference to it and
        # discard this Program object
        self.kernel = KernelHub(self)

    def set_constant_array(
        self, queue: Queue, name: str, arr: Array | numpy.ndarray[Any, numpy.dtype[Any]]
    ) -> None:
        """
        Uploads a constant array to the context's devices (**CUDA only**).

        :param queue: the queue to use for the transfer.
        :param name: the name of the constant array symbol in the code.
        :param arr: either a device or a host array.
        """
        _check_set_constant_array(queue, self.devices)
        self._sd_programs[queue.device].set_constant_array(queue, name, arr)


class KernelHub:
    """An object providing access to the host program's kernels."""

    def __init__(self, program: Program):
        self._program_ref = weakref.proxy(program)

    def __getattr__(self, kernel_name: str) -> Kernel:
        """
        Returns a :py:class:`~grunnur.program.Kernel` object for a function (CUDA)/kernel (OpenCL)
        with the name ``kernel_name``.
        """
        program = self._program_ref
        sd_kernel_adapters = {
            device: sd_program.get_kernel_adapter(kernel_name)
            for device, sd_program in program._sd_programs.items()  # noqa: SLF001
        }
        return Kernel(program, sd_kernel_adapters)


def extract_arg(
    arg: Mapping[BoundDevice, Array | Buffer | numpy.generic]
    | MultiArray
    | Array
    | Buffer
    | numpy.generic,
    device: BoundDevice,
) -> BufferAdapter | numpy.generic:
    single_device_arg: Array | Buffer | numpy.generic
    if isinstance(arg, Mapping):
        single_device_arg = arg[device]
    elif isinstance(arg, MultiArray):
        single_device_arg = arg.subarrays[device]
    else:
        single_device_arg = arg

    if isinstance(single_device_arg, Array):
        return single_device_arg.data._buffer_adapter  # noqa: SLF001
    if isinstance(single_device_arg, Buffer):
        return single_device_arg._buffer_adapter  # noqa: SLF001
    return single_device_arg


class PreparedKernel:
    """
    A kernel specialized for execution on a set of devices
    with all possible preparations and checks performed.
    """

    def __init__(
        self,
        devices: BoundMultiDevice,
        sd_kernel_adapters: Mapping[BoundDevice, KernelAdapter],
        global_sizes: Mapping[BoundDevice, Sequence[int]],
        local_sizes: Mapping[BoundDevice, Sequence[int] | None],
        hold_reference: Kernel | None = None,
    ):
        # If this object can be used by itself (e.g. when created from `Kernel.prepare()`),
        # this attribute will hold thre reference to the original `Kernel`.
        # On the other hand, in `StaticKernel` the object is used internally,
        # and holding a reference to the parent `StaticKernel` here
        # will result in a reference cycle. So `StaticKernel` will just pass `None`.
        self._hold_reference = hold_reference

        self._prepared_kernel_adapters = {}

        for device in sd_kernel_adapters:
            kernel_ls = local_sizes[device]
            kernel_gs = global_sizes[device]
            pkernel = sd_kernel_adapters[device].prepare(kernel_gs, kernel_ls)

            self._prepared_kernel_adapters[device] = pkernel

        self._devices = devices

    def __call__(
        self,
        queue: Queue | MultiQueue,
        *args: MultiArray | Array | Buffer | numpy.generic,
        local_mem: int = 0,
    ) -> Any:
        """
        Enqueues the kernel on the devices in the given queue.
        The kernel must have been prepared for all of these devices.

        If an argument is a :py:class:`~grunnur.Array` or :py:class:`~grunnur.Buffer` object,
        it must belong to the device on which the kernel is being executed
        (so ``queue`` must only have one device).

        If an argument is a :py:class:`~grunnur.MultiArray`, it should have subarrays
        on all the devices from the given ``queue``.

        If an argument is a ``numpy`` scalar, it will be passed to the kernel directly.

        If an argument is a integer-keyed ``dict``, its values corresponding to the
        device indices the kernel is executed on will be passed as kernel arguments.

        :param args: kernel arguments.
        :param kwds: backend-specific keyword parameters.
        :returns: a list of ``Event`` objects for enqueued kernels in case of PyOpenCL.
        """
        if isinstance(queue, Queue):
            queue = MultiQueue([queue])

        # Technically this would be caught by `issubset()`, but it'll help to provide
        # a more specific error to the user.
        if queue.devices.context != self._devices.context:
            raise ValueError("The provided queue must belong to the same context this program uses")

        if not queue.devices.issubset(self._devices):
            raise ValueError(
                f"Requested execution on devices {queue.devices}; "
                f"only compiled for {self._devices}"
            )

        ret_vals = []
        for device in queue.devices:
            kernel_args = [extract_arg(arg, device) for arg in args]

            single_queue = queue.queues[device]

            pkernel = self._prepared_kernel_adapters[device]
            ret_val = pkernel(single_queue._queue_adapter, *kernel_args, local_mem=local_mem)  # noqa: SLF001
            ret_vals.append(ret_val)

        return ret_vals


def normalize_sizes(
    devices: Sequence[BoundDevice],
    global_size: Sequence[int] | Mapping[BoundDevice, Sequence[int]],
    local_size: Sequence[int] | None | Mapping[BoundDevice, Sequence[int] | None] = None,
) -> tuple[
    BoundMultiDevice,
    dict[BoundDevice, tuple[int, ...]],
    dict[BoundDevice, tuple[int, ...] | None],
]:
    if not isinstance(global_size, Mapping):
        global_size = {device: global_size for device in devices}

    if not isinstance(local_size, Mapping):
        local_size = {device: local_size for device in devices}

    normalized_global_size = {device: tuple(gs) for device, gs in global_size.items()}
    normalized_local_size = {
        device: tuple(ls) if ls is not None else None for device, ls in local_size.items()
    }

    if normalized_global_size.keys() != normalized_local_size.keys():
        raise ValueError(
            "Mismatched device sets for global and local sizes: "
            f"local sizes have {list(normalized_local_size.keys())}, "
            f"global sizes have {list(normalized_global_size.keys())}"
        )

    devices_subset = BoundMultiDevice.from_bound_devices(
        [device for device in devices if device in normalized_global_size]
    )

    return devices_subset, normalized_global_size, normalized_local_size


class Kernel:
    """A kernel compiled for multiple devices."""

    def __init__(self, program: Program, sd_kernel_adapters: dict[BoundDevice, KernelAdapter]):
        self._program = program
        self._sd_kernel_adapters = sd_kernel_adapters

    @property
    def max_total_local_sizes(self) -> dict[BoundDevice, int]:
        """
        The maximum possible number of threads in a block (CUDA)/work items in a work group (OpenCL)
        for this kernel.
        """
        return {
            device: sd_kernel_adapter.max_total_local_size
            for device, sd_kernel_adapter in self._sd_kernel_adapters.items()
        }

    def prepare(
        self,
        global_size: Sequence[int] | Mapping[BoundDevice, Sequence[int]],
        local_size: Sequence[int] | None | Mapping[BoundDevice, Sequence[int] | None] = None,
    ) -> PreparedKernel:
        """
        Prepares the kernel for execution.

        If ``local_size`` or ``global_size`` are integer, they will be treated as 1-tuples.

        One can pass specific global and local sizes for each device
        using dictionaries keyed with device indices.
        This achieves another purpose: the kernel will only be prepared for those devices,
        and not for all devices available in the context.

        :param global_size: the total number of threads (CUDA)/work items (OpenCL) in each dimension
            (column-major). Note that there may be a maximum size in each dimension as well
            as the maximum number of dimensions.
            See :py:class:`~grunnur.adapter_base.DeviceParameters` for details.
        :param local_size: the number of threads in a block (CUDA)/work items in a
            work group (OpenCL) in each dimension (column-major).
            If ``None``, it will be chosen automatically.
        """
        multi_device, n_global_size, n_local_size = normalize_sizes(
            self._program.devices, global_size, local_size
        )

        # Filter out only the kernel adapters mentioned in global/local_size
        sd_kernel_adapters = {device: self._sd_kernel_adapters[device] for device in multi_device}

        return PreparedKernel(
            multi_device, sd_kernel_adapters, n_global_size, n_local_size, hold_reference=self
        )

    def __call__(
        self,
        queue: Queue | MultiQueue,
        global_size: Sequence[int] | Mapping[BoundDevice, Sequence[int]],
        local_size: Sequence[int] | None | Mapping[BoundDevice, Sequence[int] | None] = None,
        *args: MultiArray | Array | Buffer | numpy.generic,
        local_mem: int = 0,
    ) -> Any:
        """
        A shortcut for :py:meth:`Kernel.prepare` and subsequent :py:meth:`PreparedKernel.__call__`.
        See their doc entries for details.
        """
        pkernel = self.prepare(global_size, local_size)
        return pkernel(queue, *args, local_mem=local_mem)



from __future__ import annotations

import inspect
import os.path
import re
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Callable, Iterable, Mapping, Sequence

import mako.exceptions
import mako.template


class RenderError(Exception):
    """A custom wrapper for Mako template render errors, to facilitate debugging."""

    exception: Exception
    """The original exception thrown by Mako's `render()`."""

    args: tuple[Any, ...]
    """The arguments used to render the template."""

    globals: dict[str, Any]
    """The globals used to render the template."""

    source: str
    """The source of the template."""

    def __init__(
        self,
        exception: Exception,
        args: Sequence[Any],
        globals_: Mapping[str, Any],
        source: str,
        mako_traceback: str,
    ):
        super().__init__()
        self.exception = exception
        self.args = tuple(args)
        self.globals = dict(globals_)
        self.source = source
        self.mako_traceback = mako_traceback

    def __str__(self) -> str:
        return (
            "Failed to render a template with\n"
            f"* args: {self.args}\n* globals: {self.globals}\n* source:\n{self.source}\n"
            f"* Mako error: ({type(self.exception).__name__}) {self.exception}\n\n"
            f"* Mako traceback:{self.mako_traceback}"
        )


def _extract_def_source(source: str, name: str) -> str:
    """
    Attempts to extract the source of a single def from Mako template.
    This makes error messages much more readable.
    """
    match = re.search(
        r"(<%def\s+name\s*=\s*[\"']" + name + r"\(.*?>.*?</%def>)", source, flags=re.DOTALL
    )
    if not match:
        warnings.warn(
            f"Could not find the template definition '{name}'", SyntaxWarning, stacklevel=1
        )
        return source

    return match.group(1)


def _make_template(filename: Path | None = None, text: str | None = None) -> mako.template.Template:
    return mako.template.Template(
        text=text,
        filename=str(filename) if filename else None,
        strict_undefined=True,
        imports=["import numpy"],
    )


class Template:
    """A wrapper for mako ``Template`` objects."""

    @classmethod
    def from_associated_file(cls, filename: str) -> Template:
        """
        Returns a :py:class:`Template` object created from the file
        which has the same name as ``filename`` and the extension ``.mako``.
        Typically used in computation modules as ``Template.from_associated_file(__file__)``.
        """
        template_path = Path(filename).resolve().with_suffix(".mako")
        mako_template = _make_template(filename=template_path)
        return cls(mako_template)

    @classmethod
    def from_string(cls, template_source: str) -> Template:
        """Returns a :py:class:`Template` object created from source."""
        mako_template = _make_template(text=template_source)
        return cls(mako_template)

    def __init__(self, mako_template: mako.template.Template):
        self._mako_template = mako_template
        self._defs: dict[str, DefTemplate] = {}

    def get_def(self, name: str) -> DefTemplate:
        """Returns the template def with the name ``name``."""
        if name in self._defs:
            return self._defs[name]

        if name not in self._mako_template.list_defs():
            raise AttributeError(f"Template has no def '{name}'")

        def_source = _extract_def_source(self._mako_template.source, name)
        def_template = DefTemplate(name, self._mako_template.get_def(name), def_source)
        self._defs[name] = def_template
        return def_template


class DefTemplate:
    """A wrapper for Mako ``DefTemplate`` objects."""

    @classmethod
    def from_callable(cls, name: str, callable_obj: Callable[..., str]) -> DefTemplate:
        """
        Creates a template def from a callable returning a string.
        The parameter list of the callable is used to create the pararameter list
        of the resulting template def; the callable should return the body of a
        Mako template def regardless of the arguments it receives.
        """
        signature = inspect.signature(callable_obj)

        # pass mock values to extract the value
        args = [None] * len(signature.parameters)

        return cls._from_signature_and_body(name, signature, callable_obj(*args))

    @classmethod
    def from_string(cls, name: str, argnames: Iterable[str], source: str) -> DefTemplate:
        """Creates a template def from a string with its body and a list of argument names."""
        kind = inspect.Parameter.POSITIONAL_OR_KEYWORD
        parameters = [inspect.Parameter(name, kind=kind) for name in argnames]
        signature = inspect.Signature(parameters)

        return cls._from_signature_and_body(name, signature, source)

    @classmethod
    def _from_signature_and_body(
        cls, name: str, signature: inspect.Signature, body: str
    ) -> DefTemplate:
        src = "<%def name='" + name + str(signature) + "'>\n" + body + "\n</%def>"
        mako_def_template = _make_template(text=src).get_def(name)
        return cls(name, mako_def_template, src)

    def __init__(self, name: str, mako_def_template: mako.template.DefTemplate, source: str):
        self.name = name
        self._mako_def_template = mako_def_template
        self.source = source

    def render(self, *args: Any, **globals_: Any) -> str:
        """Renders the template def with given arguments and globals."""
        try:
            return self._mako_def_template.render(*args, **globals_)
        except RenderError:
            # _render() can be called by a chain of templates which call each other;
            # passing the original render error to the top so that it could be handled there.
            raise
        except Exception as exc:
            mako_traceback = mako.exceptions.text_error_template().render()
            raise RenderError(exc, args, globals_, self.source, mako_traceback) from exc



import re

import pytest

from grunnur.utils import (
    all_different,
    all_same,
    bounding_power_of_2,
    find_local_size,
    get_launch_size,
    log2,
    max_factor,
    min_blocks,
    normalize_object_sequence,
    prod,
    string_matches_masks,
    update_dict,
)


def test_all_same():
    assert not all_same([1, 1, 3])
    assert all_same([1, 1, 1])


def test_all_different():
    assert not all_different([1, 1, 3])
    assert all_different([1, 2, 3])


def test_min_blocls():
    assert min_blocks(19, 10) == 2
    assert min_blocks(20, 10) == 2
    assert min_blocks(21, 10) == 3
    with pytest.raises(ValueError, match="The length must be positive"):
        assert min_blocks(0, 10)


def test_log2():
    assert log2(1) == 0
    assert log2(2) == 1
    assert log2(3) == 1
    assert log2(4) == 2


def test_bounding_power_of_2():
    assert bounding_power_of_2(1) == 1
    assert bounding_power_of_2(2) == 2
    assert bounding_power_of_2(3) == 4
    assert bounding_power_of_2(4) == 4
    assert bounding_power_of_2(5) == 8


def test_prod():
    assert prod([]) == 1
    assert prod([1, 2, 3]) == 6


def test_string_matches_masks():
    assert string_matches_masks("foo")
    assert string_matches_masks("foo", include_masks=["fo", "ba"])
    assert not string_matches_masks("foo", include_masks=["ff", "ba"])
    assert not string_matches_masks("foo", exclude_masks=["fo", "ba"])
    assert string_matches_masks("foo", exclude_masks=["ff", "ba"])


def test_normalize_object_sequence():
    assert normalize_object_sequence([1], int) == (1,)
    assert normalize_object_sequence([1, 2], int) == (1, 2)

    # Empty sequence
    with pytest.raises(
        ValueError, match="The iterable of base objects for the context cannot be empty"
    ):
        normalize_object_sequence([], str)

    # Some of the objects are equal
    with pytest.raises(ValueError, match="All base objects must be different"):
        normalize_object_sequence([1, 2, 1], int)

    # Wrong type
    with pytest.raises(
        TypeError,
        match=re.escape(
            "The iterable must contain only subclasses of <class 'str'>, "
            "got [<class 'str'>, <class 'int'>]"
        ),
    ):
        normalize_object_sequence(["1", 2], str)


def test_max_factor():
    assert max_factor(3, 6) == 3
    assert max_factor(6, 6) == 6
    assert max_factor(37 * 37, 1024) == 37
    assert max_factor(512 * 37, 1024) == 37 * 16
    assert max_factor(1024 * 37, 1024) == 1024
    assert max_factor(521, 512) == 1


def test_find_local_size():
    assert find_local_size((10 * 127, 3 * 127, 300), (64, 64, 32), 64) == (10, 3, 2)
    assert find_local_size((10 * 127, 3 * 127, 300), (1, 1, 1), 1) == (1, 1, 1)


def test_get_launch_size():
    with pytest.raises(ValueError, match="Global size has too many dimensions"):
        get_launch_size((64, 64, 64), 64, (100, 100, 100, 100))

    with pytest.raises(ValueError, match="Global/local work sizes have differing dimensions"):
        get_launch_size((64, 64, 64), 64, (100, 100, 100), (10, 10))

    with pytest.raises(
        ValueError, match="Global sizes must be multiples of corresponding local sizes"
    ):
        get_launch_size((64, 64, 64), 64, (100, 100, 100), (10, 10, 15))

    assert get_launch_size((64, 64, 32), 64, (10 * 127, 3 * 127, 300)) == (
        (127, 127, 150),
        (10, 3, 2),
    )
    assert get_launch_size((64, 64, 32), 64, (100, 200, 300), (5, 10, 1)) == (
        (20, 20, 300),
        (5, 10, 1),
    )


def test_update_dict():
    assert update_dict({1: 2, 2: 3}, {3: 4}) == {1: 2, 2: 3, 3: 4}
    with pytest.raises(
        ValueError, match="Cannot add an item '2' - it already exists in the old dictionary"
    ):
        update_dict({1: 2, 2: 3}, {2: 4})
    with pytest.raises(ValueError, match="Custom error message for '2'"):
        update_dict({1: 2, 2: 3}, {2: 4}, error_msg="Custom error message for '{name}'")



from collections.abc import Sequence
from typing import Any

class Template:
    source: str
    def __init__(
        self,
        *,
        text: str | None = None,
        filename: str | None = None,
        strict_undefined: bool = False,
        imports: Sequence[str] | None = None,
    ): ...
    def get_def(self, name: str) -> DefTemplate: ...
    def list_defs(self) -> list[str]: ...

class DefTemplate:
    def render(self, *args: Any, **kwargs: Any) -> str: ...



import pytest

from grunnur import API, Platform, PlatformFilter, cuda_api_id, opencl_api_id


def test_all(mock_backend_pyopencl):
    mock_backend_pyopencl.add_platform_with_devices("Platform1", ["Device1"])
    mock_backend_pyopencl.add_platform_with_devices("Platform2", ["Device2"])
    api = API.from_api_id(mock_backend_pyopencl.api_id)
    platforms = Platform.all(api)
    platform_names = {platform.name for platform in platforms}
    assert platform_names == {"Platform1", "Platform2"}


def test_all_filtered(mock_backend_pyopencl):
    mock_backend_pyopencl.add_platform_with_devices("foo-bar", ["Device1"])
    mock_backend_pyopencl.add_platform_with_devices("bar-baz", ["Device2"])
    mock_backend_pyopencl.add_platform_with_devices("foo-baz", ["Device3"])
    api = API.from_api_id(mock_backend_pyopencl.api_id)
    platforms = Platform.all_filtered(
        api, PlatformFilter(include_masks=["foo"], exclude_masks=["bar"])
    )
    assert len(platforms) == 1
    assert platforms[0].name == "foo-baz"


def test_from_backend_platform(mock_backend_factory):
    mock_backend_pyopencl = mock_backend_factory.mock(opencl_api_id())
    mock_backend_pyopencl.add_platform_with_devices("Platform1", ["Device1"])
    mock_backend_pyopencl.add_platform_with_devices("Platform2", ["Device2"])

    # Add a CUDA API to make sure it is queried whether the object is its platform object
    # (even though CUDA doesn't have platforms)
    mock_backend_pycuda = mock_backend_factory.mock(cuda_api_id())
    mock_backend_pycuda.add_devices(["Device1", "Device2"])

    backend_platform = mock_backend_pyopencl.pyopencl.get_platforms()[0]

    with pytest.raises(TypeError, match="was not recognized as a platform object"):
        Platform.from_backend_platform(1)

    platform = Platform.from_backend_platform(backend_platform)
    assert platform.api.id == opencl_api_id()
    assert platform.name == "Platform1"


def test_from_index(mock_backend_pyopencl):
    mock_backend_pyopencl.add_platform_with_devices("Platform1", ["Device1"])
    mock_backend_pyopencl.add_platform_with_devices("Platform2", ["Device2"])
    api = API.from_api_id(mock_backend_pyopencl.api_id)

    platform = Platform.from_index(api, 1)
    assert platform.name == "Platform2"


def test_eq(mock_backend):
    if mock_backend.api_id == cuda_api_id():
        mock_backend.add_devices(["Device1"])
        api = API.from_api_id(mock_backend.api_id)

        p0_v1 = Platform.from_index(api, 0)
        p0_v2 = Platform.from_index(api, 0)

        assert p0_v1 is not p0_v2
        assert p0_v1 == p0_v2

    else:
        mock_backend.add_platform_with_devices("Platform0", ["Device0"])
        mock_backend.add_platform_with_devices("Platform1", ["Device1"])
        api = API.from_api_id(mock_backend.api_id)

        p0_v1 = Platform.from_index(api, 0)
        p0_v2 = Platform.from_index(api, 0)
        p1 = Platform.from_index(api, 1)

        assert p0_v1 is not p0_v2
        assert p0_v1 == p0_v2
        assert p0_v1 != p1


def test_hash(mock_backend):
    if mock_backend.api_id == cuda_api_id():
        mock_backend.add_devices(["Device1"])
        api = API.from_api_id(mock_backend.api_id)

        p0 = Platform.from_index(api, 0)

        d = {p0: 0}
        assert d[p0] == 0

    else:
        mock_backend.add_platform_with_devices("Platform0", ["Device0"])
        mock_backend.add_platform_with_devices("Platform1", ["Device1"])
        api = API.from_api_id(mock_backend.api_id)

        p0 = Platform.from_index(api, 0)
        p1 = Platform.from_index(api, 1)

        d = {p0: 0, p1: 1}
        assert d[p0] == 0
        assert d[p1] == 1


def test_getitem(mock_backend_pyopencl):
    mock_backend_pyopencl.add_platform_with_devices("Platform0", ["Device0"])
    mock_backend_pyopencl.add_platform_with_devices("Platform1", ["Device1", "Device2"])
    api = API.from_api_id(mock_backend_pyopencl.api_id)

    p1 = Platform.from_index(api, 1)

    assert p1.devices[0].name == "Device1"
    assert p1.devices[1].name == "Device2"


def test_attributes(mock_backend):
    mock_backend.add_devices(["Device1"])
    api = API.from_api_id(mock_backend.api_id)
    p = Platform.from_index(api, 0)

    assert p.api == api
    assert p.name == {cuda_api_id(): "nVidia CUDA", opencl_api_id(): "Platform0"}[api.id]
    assert p.shortcut == api.shortcut + ",0"
    assert str(p) == "platform(" + p.shortcut + ")"
    assert p.vendor == {cuda_api_id(): "nVidia", opencl_api_id(): "Mock Platforms"}[api.id]
    assert p.version == {cuda_api_id(): "CUDA 10.0.0", opencl_api_id(): "OpenCL 1.2"}[api.id]



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .adapter_cuda import CuAPIAdapterFactory
from .adapter_opencl import OclAPIAdapterFactory

if TYPE_CHECKING:  # pragma: no cover
    from .adapter_base import APIID, APIAdapter
    from .platform import Platform


def cuda_api_id() -> APIID:
    """Returns the identifier of CUDA API."""
    return CuAPIAdapterFactory().api_id


def opencl_api_id() -> APIID:
    """Returns the identifier of OpenCL API."""
    return OclAPIAdapterFactory().api_id


_ALL_API_ADAPTER_FACTORIES = {
    factory.api_id: factory
    for factory in [
        CuAPIAdapterFactory(),
        OclAPIAdapterFactory(),
    ]
}


def all_api_ids() -> list[APIID]:
    """Returns a list of identifiers for all APIs available."""
    return list(_ALL_API_ADAPTER_FACTORIES.keys())


class API:
    """A generalized GPGPU API."""

    id: APIID
    """This API's ID."""

    shortcut: str
    """
    A shortcut for this API (to use in :py:meth:`all_by_shortcut`,
    usually coming from some kind of a CLI).
    Equal to ``id.shortcut``.
    """

    @classmethod
    def all_available(cls) -> list[API]:
        """
        Returns a list of :py:class:`~grunnur.API` objects
        for which backends are available.
        """
        return [
            cls.from_api_id(api_id)
            for api_id, api_factory in _ALL_API_ADAPTER_FACTORIES.items()
            if api_factory.available
        ]

    @classmethod
    def all_by_shortcut(cls, shortcut: str | None = None) -> list[API]:
        """
        If ``shortcut`` is a string, returns a list of one :py:class:`~grunnur.API` object
        whose :py:attr:`~API.id` attribute has its
        :py:attr:`~grunnur.adapter_base.APIID.shortcut` attribute equal to it
        (or raises an error if it was not found, or its backend is not available).

        If ``shortcut`` is ``None``, returns a list of all available
        :py:class:`~grunnur.API` objects.

        :param shortcut: an API shortcut to match.
        """
        if shortcut is None:
            apis = cls.all_available()
        else:
            for api_id, api_factory in _ALL_API_ADAPTER_FACTORIES.items():
                if shortcut == api_id.shortcut:
                    if not api_factory.available:
                        raise ValueError(str(shortcut) + " API is not available")
                    apis = [cls.from_api_id(api_id)]
                    break
            else:
                raise ValueError("Invalid API shortcut: " + str(shortcut))

        return apis

    @classmethod
    def from_api_id(cls, api_id: APIID) -> API:
        """
        Creates an :py:class:`~grunnur.API` object out of an identifier.

        :param api_id: API identifier.
        """
        api_adapter = _ALL_API_ADAPTER_FACTORIES[api_id].make_api_adapter()
        return cls(api_adapter)

    def __init__(self, api_adapter: APIAdapter):
        self._api_adapter = api_adapter
        self.id = api_adapter.id
        self.shortcut = self.id.shortcut

    @property
    def platforms(self) -> list[Platform]:
        """A list of this API's :py:class:`Platform` objects."""
        from .platform import Platform  # avoiding circular imports

        return Platform.all(self)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, API) and self._api_adapter == other._api_adapter

    def __hash__(self) -> int:
        return hash((type(self), self._api_adapter))

    def __str__(self) -> str:
        return f"api({self.shortcut})"



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol, runtime_checkable

from .dtypes import _normalize_type

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Sequence

    import numpy
    from numpy.typing import DTypeLike


@runtime_checkable
class ArrayMetadataLike(Protocol):
    """
    A protocol for an object providing array metadata.
    :py:class:`numpy.ndarray` or :py:class:`Array` follow this protocol.
    """

    @property
    def shape(self) -> tuple[int, ...]:
        """Array shape."""

    @property
    def dtype(self) -> numpy.dtype[Any]:
        """The type of an array element."""


class ArrayMetadata:
    """
    A helper object for array-like classes that handles shape/strides/buffer size checks
    without actual data attached to it.
    """

    shape: tuple[int, ...]
    """Array shape."""

    dtype: numpy.dtype[Any]
    """Array item data type."""

    strides: tuple[int, ...]
    """Array strides."""

    is_contiguous: bool
    """If ``True``, means that array's data forms a continuous chunk of memory."""

    @classmethod
    def from_arraylike(cls, array: ArrayMetadataLike) -> ArrayMetadata:
        return cls(array.shape, array.dtype, strides=getattr(array, "strides", None))

    def __init__(
        self,
        shape: Sequence[int],
        dtype: DTypeLike,
        strides: Sequence[int] | None = None,
        first_element_offset: int = 0,
        buffer_size: int | None = None,
    ):
        shape = tuple(shape)
        dtype = _normalize_type(dtype)

        if len(shape) == 0:
            raise ValueError("Array shape cannot be an empty sequence")

        default_strides = _get_strides(shape, dtype.itemsize)

        if strides is None:
            strides = default_strides
            self.contiguous = True
        else:
            strides = tuple(strides)
            # Technically, an array with non-default (e.g., overlapping) strides
            # can be contioguous, but that's too hard to determine.
            self.contiguous = strides == default_strides

        min_offset, max_offset = _get_range(shape, dtype.itemsize, strides)
        if buffer_size is None:
            buffer_size = first_element_offset + max_offset

        full_min_offset = first_element_offset + min_offset
        if full_min_offset < 0 or full_min_offset + dtype.itemsize > buffer_size:
            raise ValueError(
                f"The minimum offset for given strides ({full_min_offset}) "
                f"is outside the given buffer range ({buffer_size})"
            )

        full_max_offset = first_element_offset + max_offset
        if full_max_offset > buffer_size:
            raise ValueError(
                f"The maximum offset for given strides ({full_max_offset}) "
                f"is outside the given buffer range ({buffer_size})"
            )

        self.shape = shape
        self.dtype = dtype
        self.strides = strides
        self.first_element_offset = first_element_offset
        self._full_min_offset = full_min_offset
        self._full_max_offset = full_max_offset
        self.buffer_size = buffer_size

    def minimal_subregion(self) -> tuple[int, int, ArrayMetadata]:
        """
        Returns the metadata for the minimal subregion that fits all the data in this view,
        along with the subgregion offset in the current buffer and the required subregion length.
        """
        subregion_origin = self._full_min_offset
        subregion_size = self._full_max_offset - self._full_min_offset
        new_metadata = ArrayMetadata(
            self.shape,
            self.dtype,
            strides=self.strides,
            first_element_offset=self.first_element_offset - self._full_min_offset,
            buffer_size=subregion_size,
        )
        return subregion_origin, subregion_size, new_metadata

    def __getitem__(self, slices: slice | tuple[slice, ...]) -> ArrayMetadata:
        if isinstance(slices, slice):
            slices = (slices,)
        if len(slices) < len(self.shape):
            slices += (slice(None),) * (len(self.shape) - len(slices))
        new_fe_offset, new_shape, new_strides = _get_view(self.shape, self.strides, slices)
        return ArrayMetadata(
            new_shape, self.dtype, strides=new_strides, first_element_offset=new_fe_offset
        )


def _get_strides(shape: Sequence[int], itemsize: int) -> tuple[int, ...]:
    # Constructs strides for a contiguous array of shape ``shape`` and item size ``itemsize``.
    strides = []
    stride = itemsize
    for length in reversed(shape):
        strides.append(stride)
        stride *= length
    return tuple(reversed(strides))


def _normalize_slice(length: int, stride: int, slice_: slice) -> tuple[int, int, int]:
    """
    Given a slice over an array of length ``length`` with the stride ``stride`` between elements,
    return a tuple ``(offset, last, stride)`` where ``offset`` is the offset of the first
    element of the resulting view, ``last`` is the index of the last element of the view (0-based),
    and ``stride`` is the new stride between elements.
    """
    start, stop, step = slice_.indices(length)

    offset = start * stride
    total_elems = abs(stop - start)
    length = (total_elems - 1) // abs(step) + 1
    new_stride = stride * step

    return offset, length, new_stride


def _get_view(
    shape: Sequence[int], strides: Sequence[int], slices: Sequence[slice]
) -> tuple[int, tuple[int, ...], tuple[int, ...]]:
    """
    Given an array shape and strides, and a sequence of slices defining a view,
    returns a tuple of three elements: the offset of the first element of the view,
    the view shape and the view strides.
    """
    if len(strides) != len(shape):
        raise ValueError("Shape and strides must have the same length")
    if len(slices) != len(shape):
        raise ValueError("Shape and slices must have the same length")

    offsets, lengths, strides = zip(
        *[
            _normalize_slice(length, stride, slice_)
            for length, stride, slice_ in zip(shape, strides, slices, strict=True)
        ],
        strict=True,
    )

    return sum(offsets), tuple(lengths), tuple(strides)


def _get_range(shape: Sequence[int], itemsize: int, strides: Sequence[int]) -> tuple[int, int]:
    """
    Given an array shape, item size (in bytes), and a sequence of strides,
    returns a pair ``(min_offset, max_offset)``,
    where ``min_offset`` is the minimum byte offset of an array element,
    and ``max_offset`` is the maximum byte offset of an array element plus itemsize.
    """
    if len(strides) != len(shape):
        raise ValueError("Shape and strides must have the same length")

    # Now the address of an element (i1, i2, ...) of the resulting view is
    #     addr = i1 * stride1 + i2 * stride2 + ...,
    #     where 0 <= i_k <= length_k - 1
    # We want to find the minimum and the maximum value of addr,
    # keeping in mind that strides may be negative.
    # Since it is a linear function of each index, the extrema will be located
    # at the ends of intervals, so we can find minima and maxima for each term separately.

    # Since we separated the offsets already, for each dimension the address
    # of the first element is 0. We calculate the address of the last byte in each dimension.
    last_addrs = [(length - 1) * stride for length, stride in zip(shape, strides, strict=True)]

    # Sort the pairs (0, last_addr)
    pairs = [(0, last_addr) if last_addr > 0 else (last_addr, 0) for last_addr in last_addrs]
    minima, maxima = zip(*pairs, strict=True)

    min_offset = sum(minima)
    max_offset = sum(maxima) + itemsize

    return min_offset, max_offset



"""
:py:class:`~grunnur.Module` factories
which are used to compensate for the lack of complex number operations in OpenCL,
and the lack of C++ synthax which would allow one to write them.
"""

from __future__ import annotations

from typing import Any
from warnings import warn

import numpy

from . import dtypes
from .modules import Module
from .template import Template

TEMPLATE = Template.from_associated_file(__file__)


def check_information_loss(out_dtype: numpy.dtype[Any], expected_dtype: numpy.dtype[Any]) -> None:
    if dtypes.is_complex(expected_dtype) and not dtypes.is_complex(out_dtype):
        warn(
            "Imaginary part ignored during the downcast from "
            + str(expected_dtype)
            + " to "
            + str(out_dtype),
            numpy.exceptions.ComplexWarning,
            stacklevel=2,
        )


def derive_out_dtype(
    *in_dtypes: numpy.dtype[Any], out_dtype: numpy.dtype[Any] | None = None
) -> numpy.dtype[Any]:
    expected_dtype = dtypes.result_type(*in_dtypes)
    if out_dtype is None:
        out_dtype = expected_dtype
    else:
        check_information_loss(out_dtype, expected_dtype)
    return out_dtype


def cast(in_dtype: numpy.dtype[Any], out_dtype: numpy.dtype[Any]) -> Module:
    """
    Returns a :py:class:`~grunnur.Module` with a function of one argument
    that casts values of ``in_dtype`` to ``out_dtype``.
    """
    upcast_to_complex = not dtypes.is_complex(in_dtype) and dtypes.is_complex(out_dtype)
    same_space = dtypes.is_complex(out_dtype) == dtypes.is_complex(in_dtype)
    if not upcast_to_complex and not same_space:
        raise ValueError(f"cast from {in_dtype} to {out_dtype} is not supported")

    return Module(
        TEMPLATE.get_def("cast"),
        render_globals=dict(
            dtypes=dtypes,
            out_dtype=out_dtype,
            in_dtype=in_dtype,
            upcast_to_complex=upcast_to_complex,
            same_space=same_space,
        ),
    )


def add(*in_dtypes: numpy.dtype[Any], out_dtype: numpy.dtype[Any] | None = None) -> Module:
    """
    Returns a :py:class:`~grunnur.Module`  with a function of
    ``len(in_dtypes)`` arguments that adds values of types ``in_dtypes``.
    If ``out_dtype`` is given, it will be set as a return type for this function.

    This is necessary since on some platforms complex numbers are based on 2-vectors,
    and therefore the ``+`` operator for a complex and a real number
    works in an unexpected way (returning ``(a.x + b, a.y + b)`` instead of ``(a.x + b, a.y)``).
    """
    out_dtype = derive_out_dtype(*in_dtypes, out_dtype=out_dtype)
    return Module(
        TEMPLATE.get_def("add_or_mul"),
        render_globals=dict(dtypes=dtypes, op="add", out_dtype=out_dtype, in_dtypes=in_dtypes),
    )


def mul(*in_dtypes: numpy.dtype[Any], out_dtype: numpy.dtype[Any] | None = None) -> Module:
    """
    Returns a :py:class:`~grunnur.Module`  with a function of
    ``len(in_dtypes)`` arguments that multiplies values of types ``in_dtypes``.
    If ``out_dtype`` is given, it will be set as a return type for this function.
    """
    out_dtype = derive_out_dtype(*in_dtypes, out_dtype=out_dtype)
    return Module(
        TEMPLATE.get_def("add_or_mul"),
        render_globals=dict(dtypes=dtypes, op="mul", out_dtype=out_dtype, in_dtypes=in_dtypes),
    )


def div(
    dividend_dtype: numpy.dtype[Any],
    divisor_dtype: numpy.dtype[Any],
    out_dtype: numpy.dtype[Any] | None = None,
) -> Module:
    """
    Returns a :py:class:`~grunnur.Module` with a function of two arguments
    that divides a value of type ``dividend_dtype`` by a value of type ``divisor_dtype``.
    If ``out_dtype`` is given, it will be set as a return type for this function.
    """
    out_dtype = derive_out_dtype(dividend_dtype, divisor_dtype, out_dtype=out_dtype)
    return Module(
        TEMPLATE.get_def("div"),
        render_globals=dict(
            dtypes=dtypes,
            out_dtype=out_dtype,
            dividend_dtype=dividend_dtype,
            divisor_dtype=divisor_dtype,
        ),
    )


def conj(dtype: numpy.dtype[Any]) -> Module:
    """
    Returns a :py:class:`~grunnur.Module` with a function of one argument
    that conjugates the value of type ``dtype``
    (if it is not a complex data type, the value will not be modified).
    """
    return Module(TEMPLATE.get_def("conj"), render_globals=dict(dtypes=dtypes, dtype=dtype))


def polar_unit(dtype: numpy.dtype[Any]) -> Module:
    """
    Returns a :py:class:`~grunnur.Module` with a function of one argument
    that returns a complex number ``exp(i * theta) == (cos(theta), sin(theta))``
    for a value ``theta`` of type ``dtype`` (must be a real data type).
    """
    if not dtypes.is_real(dtype):
        raise ValueError("polar_unit() can only be applied to real dtypes")

    return Module(TEMPLATE.get_def("polar_unit"), render_globals=dict(dtypes=dtypes, dtype=dtype))


def norm(dtype: numpy.dtype[Any]) -> Module:
    """
    Returns a :py:class:`~grunnur.Module` with a function of one argument
    that returns the 2-norm of the value of type ``dtype``
    (product by the complex conjugate if the value is complex, square otherwise).
    """
    return Module(TEMPLATE.get_def("norm"), render_globals=dict(dtypes=dtypes, dtype=dtype))


def exp(dtype: numpy.dtype[Any]) -> Module:
    """
    Returns a :py:class:`~grunnur.Module` with a function of one argument
    that exponentiates the value of type ``dtype``
    (must be a real or a complex data type).
    """
    # Supporting this will require an explicit output type specification.
    if dtypes.is_integer(dtype):
        raise ValueError(f"exp() of {dtype} is not supported")

    polar_unit_ = None if dtypes.is_real(dtype) else polar_unit(dtypes.real_for(dtype))
    return Module(
        TEMPLATE.get_def("exp"),
        render_globals=dict(dtypes=dtypes, dtype=dtype, polar_unit_=polar_unit_),
    )


def pow(  # noqa: A001
    base_dtype: numpy.dtype[Any],
    exponent_dtype: numpy.dtype[Any] | None = None,
    out_dtype: numpy.dtype[Any] | None = None,
) -> Module:
    """
    Returns a :py:class:`~grunnur.Module` with a function of two arguments
    that raises the first argument of type ``base_dtype``
    to the power of the second argument of type ``exponent_dtype``
    (an integer or real data type).

    If ``exponent_dtype`` or ``out_dtype`` are not given, they default to ``base_dtype``.
    If ``base_dtype`` is not the same as ``out_dtype``,
    the input is cast to ``out_dtype`` *before* exponentiation.
    If ``exponent_dtype`` is real, but both ``base_dtype`` and ``out_dtype`` are integer,
    a ``ValueError`` is raised.
    """
    if exponent_dtype is None:
        exponent_dtype = base_dtype

    if out_dtype is None:
        out_dtype = base_dtype

    if dtypes.is_complex(exponent_dtype):
        raise ValueError("pow() with a complex exponent is not supported")

    if dtypes.is_real(exponent_dtype):
        if dtypes.is_complex(out_dtype):
            exponent_dtype = dtypes.real_for(out_dtype)
        elif dtypes.is_real(out_dtype):
            exponent_dtype = out_dtype
        else:
            raise ValueError("pow(integer, float): integer is not supported")

    kwds = dict(
        dtypes=dtypes,
        base_dtype=base_dtype,
        exponent_dtype=exponent_dtype,
        out_dtype=out_dtype,
        div_=None,
        mul_=None,
        cast_=None,
        polar_=None,
    )
    if out_dtype != base_dtype:
        kwds["cast_"] = cast(base_dtype, out_dtype)
    if dtypes.is_integer(exponent_dtype) and not dtypes.is_real(out_dtype):
        kwds["mul_"] = mul(out_dtype, out_dtype)
        kwds["div_"] = div(out_dtype, out_dtype)
    if dtypes.is_complex(out_dtype):
        kwds["polar_"] = polar(dtypes.real_for(out_dtype))

    return Module(TEMPLATE.get_def("pow"), render_globals=kwds)


def polar(dtype: numpy.dtype[Any]) -> Module:
    """
    Returns a :py:class:`~grunnur.Module` with a function of two arguments
    that returns the complex-valued ``rho * exp(i * theta)``
    for values ``rho, theta`` of type ``dtype`` (must be a real data type).
    """
    if not dtypes.is_real(dtype):
        raise ValueError("polar() of " + str(dtype) + " is not supported")

    return Module(
        TEMPLATE.get_def("polar"),
        render_globals=dict(dtypes=dtypes, dtype=dtype, polar_unit_=polar_unit(dtype)),
    )



from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, NamedTuple

from .api import API
from .utils import string_matches_masks

if TYPE_CHECKING:  # pragma: no cover
    from .adapter_base import PlatformAdapter
    from .device import Device


class PlatformFilter(NamedTuple):
    """A set of filters for platform discovery."""

    include_masks: list[str] | None = None
    """A list of strings (treated as regexes), one of which must match the platform name."""

    exclude_masks: list[str] | None = None
    """A list of strings (treated as regexes), neither of which must match the platform name."""


class Platform:
    """A generalized GPGPU platform."""

    api: API
    """The :py:class:`~grunnur.API` object this platform belongs to."""

    name: str
    """The platform's name."""

    vendor: str
    """The platform's vendor."""

    version: str
    """The platform's version."""

    @classmethod
    def all(cls, api: API) -> list[Platform]:
        """
        Returns a list of platforms available for the given API.

        :param api: the API to search in.
        """
        return [
            Platform.from_index(api, platform_idx)
            for platform_idx in range(api._api_adapter.platform_count)  # noqa: SLF001
        ]

    @classmethod
    def all_filtered(
        cls,
        api: API,
        filter_: PlatformFilter | None = None,
    ) -> list[Platform]:
        """
        Returns a list of all platforms satisfying the given criteria in the given API.
        If ``filter`` is not provided, returns all the platforms.
        """
        if filter_ is None:
            return cls.all(api)
        return [
            platform
            for platform in cls.all(api)
            if string_matches_masks(
                platform.name,
                include_masks=filter_.include_masks,
                exclude_masks=filter_.exclude_masks,
            )
        ]

    @classmethod
    def from_backend_platform(cls, obj: Any) -> Platform:
        """Wraps a backend platform object into a Grunnur platform object."""
        for api in API.all_available():
            if api._api_adapter.isa_backend_platform(obj):  # noqa: SLF001
                platform_adapter = api._api_adapter.make_platform_adapter(obj)  # noqa: SLF001
                return cls(platform_adapter)

        raise TypeError(f"{obj} was not recognized as a platform object by any available API")

    @classmethod
    def from_index(cls, api: API, platform_idx: int) -> Platform:
        """
        Creates a platform based on its index in the list returned by the API.

        :param api: the API to search in.
        :param platform_idx: the target platform's index.
        """
        platform_adapter = api._api_adapter.get_platform_adapters()[platform_idx]  # noqa: SLF001
        return cls(platform_adapter)

    def __init__(self, platform_adapter: PlatformAdapter):
        self.api = API(platform_adapter.api_adapter)
        self._platform_adapter = platform_adapter
        self.name = self._platform_adapter.name

        self.shortcut = f"{self.api.shortcut},{platform_adapter.platform_idx}"

        self.vendor = platform_adapter.vendor
        self.version = platform_adapter.version

    @property
    def devices(self) -> list[Device]:
        """A list of this device's :py:class:`Device` objects."""
        from .device import Device  # avoiding circular imports

        return Device.all(self)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Platform) and self._platform_adapter == other._platform_adapter

    def __hash__(self) -> int:
        return hash((type(self), self._platform_adapter))

    def __str__(self) -> str:
        return f"platform({self.shortcut})"



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .api import cuda_api_id
from .context import Context
from .device import Device
from .program import (
    PreparedKernel,
    SingleDeviceProgram,
    _check_set_constant_array,
    _set_constant_array,
    normalize_sizes,
)
from .utils import prod, update_dict
from .vsize import VirtualSizeError, VirtualSizes

# the name of the global in the template containing static kernel modules
_STATIC_MODULES_GLOBAL = "static"

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Callable, Mapping, Sequence

    import numpy

    from .array import Array
    from .array_metadata import ArrayMetadataLike
    from .context import BoundDevice, BoundMultiDevice
    from .modules import Snippet
    from .queue import Queue
    from .template import DefTemplate


class StaticKernel:
    """
    An object containing a GPU kernel with fixed call sizes.

    The globals for the source template will contain an object with the name ``static``
    of the type :py:class:`~grunnur.vsize.VsizeModules` containing the id/size functions
    to be used instead of regular ones.
    """

    devices: BoundMultiDevice
    """Devices on which this kernel was compiled."""

    queue: Queue
    """The queue this static kernel was compiled and prepared for."""

    sources: dict[BoundDevice, str]
    """Source files used for each device."""

    def __init__(
        self,
        devices: Sequence[BoundDevice],
        template_src: str | Callable[..., str] | DefTemplate | Snippet,
        name: str,
        global_size: Sequence[int] | Mapping[BoundDevice, Sequence[int]],
        *,
        local_size: Sequence[int] | None | Mapping[BoundDevice, Sequence[int] | None] = None,
        render_args: Sequence[Any] = (),
        render_globals: Mapping[str, Any] = {},
        constant_arrays: Mapping[str, ArrayMetadataLike] | None = None,
        keep: bool = False,
        fast_math: bool = False,
        compiler_options: Sequence[str] | None = None,
    ):
        """
        :param devices: a single- or a multi-device object on which to compile this program.
        :param template_src: a string with the source code, or a Mako template source to render.
        :param name: the kernel's name.
        :param global_size: see :py:meth:`~grunnur.program.Kernel.prepare`.
        :param local_size: see :py:meth:`~grunnur.program.Kernel.prepare`.
        :param render_globals: a dictionary of globals to pass to the template.
        :param constant_arrays: (**CUDA only**) a dictionary ``name: (size, dtype)``
            of global constant arrays to be declared in the program.
        """
        multi_device, n_global_size, n_local_size = normalize_sizes(
            devices, global_size, local_size
        )

        self.devices = multi_device

        kernel_adapters = {}
        sources = {}
        vs_metadata = {}
        for device in multi_device:
            device_params = device.params

            kernel_ls = n_local_size[device]
            kernel_gs = n_global_size[device]

            # Since virtual size function require some registers,
            # they affect the maximum local size.
            # Start from the device's max local size as the first approximation
            # and recompile kernels with smaller local sizes until convergence.

            max_total_local_size = device_params.max_total_local_size

            while True:
                # Try to find kernel launch parameters for the requested local size.
                # May raise VirtualSizeError if it's not possible,
                # just let it pass to the caller.
                vs = VirtualSizes(
                    max_total_local_size=max_total_local_size,
                    max_local_sizes=device_params.max_local_sizes,
                    max_num_groups=device_params.max_num_groups,
                    local_size_multiple=device_params.warp_size,
                    virtual_global_size=kernel_gs,
                    virtual_local_size=kernel_ls,
                )

                new_render_globals = update_dict(
                    render_globals,
                    {_STATIC_MODULES_GLOBAL: vs.vsize_modules},
                    error_msg=(
                        f"The global name '{_STATIC_MODULES_GLOBAL}' "
                        "is reserved in static kernels"
                    ),
                )

                # Try to compile the kernel with the corresponding virtual size functions
                program = SingleDeviceProgram(
                    device,
                    template_src,
                    render_args=render_args,
                    render_globals=new_render_globals,
                    constant_arrays=constant_arrays,
                    keep=keep,
                    fast_math=fast_math,
                    compiler_options=compiler_options,
                )
                kernel_adapter = program.get_kernel_adapter(name)

                if kernel_adapter.max_total_local_size >= prod(vs.real_local_size):
                    # Kernel will execute with this local size, use it
                    break

                # By the contract of VirtualSizes, prod(vs.real_local_size) <= max_total_local_size
                # Also, since we're still in this loop,
                # kernel_adapter.max_total_local_size < prod(vs.real_local_size).
                # Therefore the new max_total_local_size value is guaranteed
                # to be smaller than the previous one.
                max_total_local_size = kernel_adapter.max_total_local_size

                # In most cases the iteration should stop at `max_total_local_size == 1`,
                # where the virtual size is trivial and always possible.
                # But occasionally we may get a kernel that cannot be executed at all
                # (e.g. it requests too much local memory),
                # and some platforms may return 0, which `VirtualSizes` will not like.
                # So we'll have a sanity check here.
                if max_total_local_size == 0:
                    raise VirtualSizeError(
                        "The kernel requires too much resourses to be executed with any local size"
                    )

            kernel_adapters[device] = kernel_adapter
            sources[device] = program.source
            vs_metadata[device] = vs

        self.sources = sources
        self._vs_metadata = vs_metadata
        self._sd_kernel_adapters = kernel_adapters

        global_sizes = {device: vs.real_global_size for device, vs in self._vs_metadata.items()}
        local_sizes = {device: vs.real_local_size for device, vs in self._vs_metadata.items()}

        self._prepared_kernel = PreparedKernel(
            multi_device, kernel_adapters, global_sizes, local_sizes
        )

    def __call__(self, queue: Queue, *args: Array | numpy.generic) -> Any:
        """
        Execute the kernel.
        In case of the OpenCL backend, returns a ``pyopencl.Event`` object.

        :param queue: the multi-device queue to use.
        :param args: kernel arguments. See :py:meth:`grunnur.program.PreparedKernel.__call__`.
        """
        return self._prepared_kernel(queue, *args)

    def set_constant_array(
        self, queue: Queue, name: str, arr: Array | numpy.ndarray[Any, numpy.dtype[Any]]
    ) -> None:
        """
        Uploads a constant array to the context's devices (**CUDA only**).

        :param queue: the queue to use for the transfer.
        :param name: the name of the constant array symbol in the code.
        :param arr: either a device or a host array.
        """
        _check_set_constant_array(queue, self.devices)
        kernel_adapter = self._sd_kernel_adapters[queue.device]
        _set_constant_array(queue, kernel_adapter.program_adapter, name, arr)
